# Copyright (c) 2026, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Simple Wait Queues
------------------

The ``drgn.helpers.linux.swait`` module provides helpers for working with
simple wait queues (``swait_queue_head`` and ``swait_queue``) from
:linux:`include/linux/swait.h`.

"""

from typing import Iterator

from drgn import Object
from drgn.helpers.linux.list import list_empty, list_for_each_entry

__all__ = (
    "swait_active",
    "swait_for_each_task",
)


def swait_active(wq: Object) -> bool:
    """
    Return whether a simple wait queue has any waiters.

    :param wq: ``struct swait_queue_head *``
    """
    return not list_empty(wq.task_list.address_of_())


def swait_for_each_task(wq: Object) -> Iterator[Object]:
    """
    Iterate over all tasks waiting on a simple wait queue.

    :param wq: ``struct swait_queue_head *``
    :return: Iterator of ``struct task_struct *`` objects.
    """
    for entry in list_for_each_entry(
        "struct swait_queue", wq.task_list.address_of_(), "task_list"
    ):
        yield entry.task



#!/usr/bin/env drgn
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Get a process core dump from drgn running against the kernel."""

import argparse
import contextlib
import functools
import io
from pathlib import Path
import struct
import sys
from typing import Iterator, List, NamedTuple, Sequence, Tuple

from drgn import (
    Architecture,
    FaultError,
    Object,
    PlatformFlags,
    Program,
    ProgramFlags,
    cast,
)
from drgn.helpers.linux.fs import d_path
from drgn.helpers.linux.list import list_for_each_entry
from drgn.helpers.linux.mm import access_remote_vm, cmdline, for_each_vma
from drgn.helpers.linux.pid import find_task

ELFCLASS32 = 1
ELFCLASS64 = 2
ELFDATA2LSB = 1
ELFDATA2MSB = 2

PT_LOAD = 1
PT_NOTE = 4
PF_X = 1 << 0
PF_W = 1 << 1
PF_R = 1 << 2

VM_READ = 0x1
VM_WRITE = 0x2
VM_EXEC = 0x4
VM_SHARED = 0x8
VM_IO = 0x4000
VM_DONTDUMP = 0x4000000


class Segment(NamedTuple):
    start: int
    end: int
    p_flags: int
    dump_size: int


class MappedFile(NamedTuple):
    path: bytes
    offset: int
    start: int
    end: int


class Phdr(NamedTuple):
    p_type: int
    p_flags: int
    p_offset: int
    p_vaddr: int
    # No p_paddr, we always set it to 0.
    p_filesz: int
    p_memsz: int
    p_align: int


def vma_snapshot(
    page_size: int, task: Object
) -> Tuple[Sequence[Segment], Sequence[MappedFile]]:
    gate_vma = task.prog_["gate_vma"].address_of_()
    special_mapping_name = task.prog_["special_mapping_name"]

    segments: List[Segment] = []
    mapped_files: List[MappedFile] = []

    def always_dump_vma(vma: Object) -> bool:
        if vma == gate_vma:
            return True
        # The kernel checks (vma->vm_ops && vma->vm_ops->name && vma->vm_ops->name(vma)).
        # As of Linux 6.9, gate_vma_name() and special_mapping_name() are the
        # only instances of ->name().
        vm_ops = vma.vm_ops.read_()
        if vm_ops:
            if vm_ops.name == special_mapping_name:
                return bool(
                    cast("struct vm_special_mapping *", vma.vm_private_data).name
                )
        return False

    def add_vma(vma: Object) -> None:
        start = vma.vm_start.value_()
        end = vma.vm_end.value_()
        flags = vma.vm_flags.read_()
        file = vma.vm_file.read_()

        p_flags = 0
        if flags & VM_READ:
            p_flags |= PF_R
        if flags & VM_WRITE:
            p_flags |= PF_W
        if flags & VM_EXEC:
            p_flags |= PF_X

        # Dumbed down version of vma_dump_size() assuming
        # (MMF_DUMP_ANON_PRIVATE|MMF_DUMP_ANON_SHARED).
        if always_dump_vma(vma):
            dump_size = end - start
        elif flags & (VM_IO | VM_DONTDUMP):
            dump_size = 0
        elif vma.anon_vma or ((flags & VM_SHARED) and file.f_inode.i_nlink == 0):
            dump_size = end - start
        elif (
            file
            and vma.vm_pgoff == 0
            and (vma.vm_flags & VM_READ)
            and file.f_inode.i_mode & 0o111
        ):
            # Include first page of executables.
            # TODO: this is out of date with Linux kernel commit 84158b7f6a06
            # ("coredump: Also dump first pages of non-executable ELF
            # libraries") (in v5.18).
            dump_size = page_size
        else:
            dump_size = 0

        if (
            segments
            and segments[-1].end == start
            and segments[-1].p_flags == p_flags
            and (
                dump_size == 0
                or segments[-1].dump_size == segments[-1].end - segments[-1].start
            )
        ):
            segments[-1] = Segment(
                start=segments[-1].start,
                end=end,
                p_flags=p_flags,
                dump_size=segments[-1].dump_size + dump_size,
            )
        else:
            segments.append(
                Segment(
                    start=start,
                    end=end,
                    p_flags=p_flags,
                    dump_size=dump_size,
                )
            )

        if file:
            path = d_path(file.f_path)
            offset = vma.vm_pgoff.value_() * page_size
            if (
                mapped_files
                and mapped_files[-1].path == path
                and mapped_files[-1].end == start
                and mapped_files[-1].offset
                + (mapped_files[-1].end - mapped_files[-1].start)
                == offset
            ):
                mapped_files[-1] = MappedFile(
                    path=path,
                    offset=mapped_files[-1].offset,
                    start=mapped_files[-1].start,
                    end=end,
                )
            else:
                mapped_files.append(
                    MappedFile(
                        path=path,
                        offset=offset,
                        start=start,
                        end=end,
                    )
                )

    for vma in for_each_vma(task.mm):
        add_vma(vma)
    add_vma(gate_vma)
    return segments, mapped_files


def _nt_pids(task: Object) -> Tuple[int, int, int, int]:
    return (
        task.pid.value_(),  # pid
        task.real_parent.pid.value_(),  # ppid
        task.tgid.value_(),  # pgrp
        task.signal.pids[prog["PIDTYPE_SID"]].numbers[0].nr.value_(),  # sid
    )


def nt_prstatus(task: Object) -> bytes:
    return struct.pack(
        "3IH2Q4I8Q27QI4x",
        0,  # info.si_signo
        0,  # info.si_code
        0,  # info.si_errno
        # TODO: can we get some of these?
        0,  # cursig
        0,  # sigpend
        0,  # sighold
        *_nt_pids(task),
        0,
        0,  # utime
        0,
        0,  # stime
        0,
        0,  # cutime
        0,
        0,  # cstime
        # reg
        *struct.unpack(
            "21Q",
            prog.read(task.stack.value_() + (4096 << 2) - 21 * 8, 21 * 8),
        ),
        0,
        0,
        0,
        0,
        0,
        0,
        # TODO: floating point registers.
        0,  # fpvalid
    )


MAX_NICE = 19
MIN_NICE = -20
NICE_WIDTH = MAX_NICE - MIN_NICE + 1
MAX_RT_PRIO = 100
DEFAULT_PRIO = MAX_RT_PRIO + NICE_WIDTH // 2


def PRIO_TO_NICE(prio: int) -> int:
    return prio - DEFAULT_PRIO


def nt_prpsinfo(task: Object, use_procfs: bool) -> bytes:
    fname_len = 16
    ELF_PRARGSZ = 80

    try:
        state: int = task.__state.value_()
    except AttributeError:
        state = task.state.value_()
    if state:
        state = (state & -state).bit_length()
    sname = ord(".") if state > 5 else b"RSDTZW"[state]

    cred = task.real_cred.read_()
    uid = cred.uid.val.value_()
    gid = cred.gid.val.value_()
    pids = _nt_pids(task)

    if use_procfs:
        psargs = (
            Path(f"/proc/{pids[0]}/cmdline")
            .read_bytes()
            .rstrip(b"\0")
            .replace(b"\0", b" ")
        )
    else:
        psargs = b" ".join(cmdline(task) or [])

    return struct.pack(
        f"4BQ6I{fname_len}s{ELF_PRARGSZ}s",
        state,
        sname,
        sname == ord("Z"),  # zomb
        PRIO_TO_NICE(task.static_prio.value_()),  # nice
        task.flags.value_(),  # flag
        uid,
        gid,
        *_nt_pids(task),
        task.comm.string_(),  # fname
        psargs,
    )


def nt_auxv(task: Object) -> bytes:
    auxv = task.mm.saved_auxv
    i = 0
    while auxv[i]:
        i += 2
    return prog.read(auxv.address_, auxv[i + 2].address_ - auxv.address_)


def nt_file(mapped_files: Sequence[MappedFile], page_size: int) -> bytes:
    buf = bytearray(16 + 24 * len(mapped_files))
    struct.pack_into("QQ", buf, 0, len(mapped_files), page_size)
    for i, mapped_file in enumerate(mapped_files):
        struct.pack_into(
            "QQQ",
            buf,
            16 + 24 * i,
            mapped_file.start,
            mapped_file.end,
            mapped_file.offset // page_size,
        )
    for mapped_file in mapped_files:
        buf.extend(mapped_file.path)
        buf.append(0)
    return buf


def gen_notes(
    task: Object, mapped_files: Sequence[MappedFile], page_size: int, use_procfs: bool
) -> bytearray:
    notes = []

    def add_nt_prstatus(t: Object) -> None:
        # This is obviously racy for the live kernel, but it's best effort.
        if t.on_cpu:
            print(f"skipping running thread {t.pid.value_()}", file=sys.stderr)
        else:
            notes.append(
                (
                    b"CORE",
                    1,  # NT_PRSTATUS
                    nt_prstatus(t),
                )
            )

    add_nt_prstatus(task)
    for t in list_for_each_entry(
        task.type_.type, task.signal.thread_head.address_of_(), "thread_node"
    ):
        if t != task:
            add_nt_prstatus(t)

    notes.append(
        (
            b"CORE",
            3,  # NT_PRPSINFO
            nt_prpsinfo(task.group_leader.read_(), use_procfs),
        )
    )
    # No NT_SIGINFO since we have no signal.
    notes.append(
        (
            b"CORE",
            6,  # NT_AUXV
            nt_auxv(task),
        )
    )
    notes.append(
        (
            b"CORE",
            0x46494C45,  # NT_FILE
            nt_file(mapped_files, page_size),
        )
    )

    buf = bytearray()
    for name, type_, desc in notes:
        buf.extend(struct.pack("III", len(name) + 1, len(desc), type_))
        buf.extend(name)
        buf.extend(bytes(4 - (len(name) & 3)))
        buf.extend(desc)
        buf.extend(bytes(-len(buf) & 3))
    return buf


def try_read_memory_procfs(
    page_size: int, mem_file: io.FileIO, address: int, size: int
) -> Iterator[Tuple[int, bytes]]:
    # An address may overflow a signed long, but we can still seek to it in
    # increments of sys.maxsize.
    whence = 0
    offset = address
    while offset:
        seek = min(offset, sys.maxsize)
        try:
            mem_file.seek(seek, whence)
        except OSError:
            # The offset returned by the lseek() system call may be negative
            # when interpreted as an off_t, which makes Python think that there
            # was an error even though the seek succeeded.
            pass
        offset -= seek
        whence = 1

    while size > 0:
        try:
            buf = mem_file.read(size)
            yield address, buf
            address += len(buf)
            size -= len(buf)
        except IOError:
            try:
                mem_file.seek(page_size, 1)
            except OSError:
                # See above.
                pass
            address += page_size
            size -= page_size


def try_read_memory_remote(
    page_size: int, mm: Object, address: int, size: int
) -> Iterator[Tuple[int, bytes]]:
    # Reading page by page isn't very efficient, but it's foolproof.
    while size > 0:
        try:
            yield address, access_remote_vm(mm, address, page_size)
        except FaultError:
            pass
        address += page_size
        size -= page_size


def main(prog: Program, argv: Sequence[str]) -> None:
    parser = argparse.ArgumentParser(
        description="Capture a process core dump without stopping it or from a kernel core dump (using drgn)"
    )
    parser.add_argument(
        "--no-procfs",
        dest="use_procfs",
        action="store_false",
        help="don't use the proc filesystem to get information about the process even when the process is local; "
        "this will skip memory that is paged out and is slower, "
        "but it can be useful if the mmap lock is deadlocked",
    )
    parser.add_argument("pid", type=int, help="PID of process to capture")
    parser.add_argument("core", type=str, help="output file")
    args = parser.parse_args(argv)

    args.use_procfs = args.use_procfs and (
        prog.flags & (ProgramFlags.IS_LIVE | ProgramFlags.IS_LOCAL)
        == (ProgramFlags.IS_LIVE | ProgramFlags.IS_LOCAL)
    )

    # TODO: these aren't necessarily the same as the kernel (e.g., when running
    # a 32-bit application on a 64-bit kernel).
    platform = prog.platform
    assert platform is not None
    ei_class = ELFCLASS64 if (platform.flags & PlatformFlags.IS_64_BIT) else ELFCLASS32
    ei_data = (
        ELFDATA2LSB
        if (platform.flags & PlatformFlags.IS_LITTLE_ENDIAN)
        else ELFDATA2MSB
    )
    if platform.arch == Architecture.X86_64:
        e_machine = 62  # EM_X86_64
    else:
        # TODO: there are assumptions that the host and target are x86-64
        # throughout this script (in struct.pack() calls, note contents).
        sys.exit("only x86-64 is supported")

    page_size = prog["PAGE_SIZE"].value_()

    ehdr_struct = struct.Struct("16BHHIQQQIHHHHHH")
    phdr_struct = struct.Struct("IIQQQQQQ")

    task = find_task(prog, args.pid)
    if not task:
        sys.exit(f"PID {args.pid} not found")

    segments, mapped_files = vma_snapshot(page_size, task)
    notes = gen_notes(task, mapped_files, page_size, args.use_procfs)

    with contextlib.ExitStack() as exit_stack:
        if args.use_procfs:
            try_read_memory = functools.partial(
                try_read_memory_procfs,
                page_size,
                exit_stack.enter_context(
                    open(f"/proc/{args.pid}/mem", "rb", buffering=0)
                ),
            )
        else:
            try_read_memory = functools.partial(
                try_read_memory_remote, page_size, task.mm.read_()
            )

        f = exit_stack.enter_context(open(args.core, "wb"))

        offset = f.seek(ehdr_struct.size)
        phdrs = [
            Phdr(
                p_type=PT_NOTE,
                p_flags=0,
                p_offset=offset,
                p_vaddr=0,
                p_filesz=len(notes),
                p_memsz=0,
                p_align=0,
            )
        ]
        f.write(notes)
        offset += len(notes)

        # Align up to a page.
        offset = f.seek(-offset % page_size, 1)
        for segment in segments:
            written_start_address = written_end_address = segment.start
            written_offset = offset
            for address, buf in try_read_memory(segment.start, segment.dump_size):
                if address == written_end_address:
                    written_end_address += len(buf)
                else:
                    phdrs.append(
                        Phdr(
                            p_type=PT_LOAD,
                            p_flags=segment.p_flags,
                            p_offset=written_offset,
                            p_vaddr=written_start_address,
                            p_filesz=written_end_address - written_start_address,
                            p_memsz=address - written_start_address,
                            p_align=page_size,
                        )
                    )
                    written_start_address = address
                    written_end_address = address + len(buf)
                    written_offset = offset
                f.write(buf)
                offset += len(buf)
            phdrs.append(
                Phdr(
                    p_type=PT_LOAD,
                    p_flags=segment.p_flags,
                    p_offset=written_offset,
                    p_vaddr=written_start_address,
                    p_filesz=written_end_address - written_start_address,
                    p_memsz=segment.end - written_start_address,
                    p_align=page_size,
                )
            )

        e_phoff = offset
        for phdr in phdrs:
            f.write(
                phdr_struct.pack(
                    phdr.p_type,
                    phdr.p_flags,
                    phdr.p_offset,
                    phdr.p_vaddr,
                    0,  # p_paddr
                    phdr.p_filesz,
                    phdr.p_memsz,
                    phdr.p_align,
                )
            )
        # TODO: >= 2**16 phdrs

        f.seek(0)
        f.write(
            ehdr_struct.pack(
                0x7F,  # ELFMAG0
                ord("E"),  # ELFMAG1
                ord("L"),  # ELFMAG2
                ord("F"),  # ELFMAG3
                ei_class,
                ei_data,
                1,  # EI_VERSION = EV_CURRENT
                0,  # EI_OSABI = ELFOSABI_NONE
                0,  # EI_ABIVERSION
                0,  # EI_PAD
                0,
                0,
                0,
                0,
                0,
                0,
                4,  # e_type = ET_CORE
                e_machine,
                1,  # e_version = EV_CURRENT
                0,  # e_entry
                e_phoff,
                0,  # e_shoff
                0,  # e_flags
                ehdr_struct.size,  # e_ehsize
                phdr_struct.size,  # e_phentsize
                len(phdrs),  # e_phnum
                0,  # e_shentsize,
                0,  # e_shnum,
                0,  # e_shstrndx
            )
        )


if __name__ == "__main__":
    prog: Program
    main(prog, sys.argv[1:])



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Swap Space
----------

The ``drgn.helpers.linux.swap`` module provides helpers for inspecting swap
partitions and swap files.
"""

from typing import Iterator

from drgn import Object, ObjectNotFoundError, PlatformFlags, Program
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.fs import d_path
from drgn.helpers.linux.mm import PageUsage
from drgn.helpers.linux.vmstat import global_node_page_state

__all__ = (
    "for_each_swap_info",
    "swap_file_path",
    "swap_is_file",
    "swap_total_usage",
    "swap_usage_in_pages",
    "total_swapcache_pages",
)


@takes_program_or_default
def for_each_swap_info(prog: Program) -> Iterator[Object]:
    """
    Iterate over all swap devices.

    :return: Iterator of ``struct swap_info_struct *`` objects.
    """
    swap_info = prog["swap_info"]
    SWP_USED = prog["SWP_USED"]
    for type in range(prog["nr_swapfiles"]):
        si = swap_info[type]
        if si.flags & SWP_USED:
            yield si


def swap_file_path(si: Object) -> bytes:
    """
    Get the path of the swap partition or file.

    >>> swap_file_path(si)
    b'/dev/sda3'

    :param si: ``struct swap_info_struct *``
    """
    return d_path(si.swap_file.f_path)


def swap_is_file(si: Object) -> bool:
    """
    Return whether a swap device is a regular file.

    :param si: ``struct swap_info_struct *``
    """
    return not (si.flags & si.prog_["SWP_BLKDEV"])


def swap_usage_in_pages(si: Object) -> int:
    """
    Get the number of pages currently in use on a swap device.

    >>> swap_usage_in_pages(si)
    394319

    :param si: ``struct swap_info_struct *``
    """
    inuse_pages = si.inuse_pages
    try:
        counter = inuse_pages.counter
    except AttributeError:
        return inuse_pages.value_()
    else:
        SWAP_USAGE_COUNTER_MASK = (
            0xBFFFFFFFFFFFFFFF
            if (
                si.prog_.platform.flags  # type: ignore[union-attr]  # platform can't be None.
                & PlatformFlags.IS_64_BIT
            )
            else 0xBFFFFFFF
        )
        return counter.value_() & SWAP_USAGE_COUNTER_MASK


@takes_program_or_default
def swap_total_usage(prog: Program) -> PageUsage:
    """
    Get the total number of swap pages and the number of free swap pages on all
    swap devices.

    >>> usage = swap_total_usage()
    >>> usage
    PageUsage(pages=2097151, free_pages=1704798)
    >>> usage.used_pages
    392353
    """
    SWP_USED = prog["SWP_USED"]
    mask = SWP_USED | prog["SWP_WRITEOK"]
    nr_to_be_unused = 0
    for si in for_each_swap_info(prog):
        if si.flags & mask == SWP_USED:
            nr_to_be_unused += swap_usage_in_pages(si)

    return PageUsage(
        pages=prog["total_swap_pages"].value_() + nr_to_be_unused,
        free_pages=prog["nr_swap_pages"].counter.value_() + nr_to_be_unused,
    )


@takes_program_or_default
def total_swapcache_pages(prog: Program) -> int:
    """
    Get the number of swap cached pages (pages that are swapped in but still
    present on a swap device).
    """
    # Since Linux kernel commit ("mm: memcg: add swapcache stat for memcg v2")
    # (in v5.12), we just have to get the NR_SWAPCACHE statistic. Before that,
    # we have to sum over the swapper spaces.
    try:
        NR_SWAPCACHE = prog["NR_SWAPCACHE"]
    except ObjectNotFoundError:
        # Since Linux kernel commit 4b3ef9daa4fc ("mm/swap: split swap cache
        # into 64MB trunks") (in v4.11), there are multiple swapper spaces per
        # swap file. Before that, there was only one per swap file.
        try:
            nr_swapper_spaces = prog["nr_swapper_spaces"]
        except ObjectNotFoundError:
            # Before Linux kernel commit 4b3ef9daa4fc ("mm/swap: split swap
            # cache into 64MB trunks") (in v4.11),
            return sum(space.nrpages.value_() for space in prog["swapper_spaces"])
        else:
            return sum(
                spaces[j].nrpages.value_()
                for nr, spaces in zip(nr_swapper_spaces, prog["swapper_spaces"])
                for j in range(nr)
            )
    else:
        return global_node_page_state(NR_SWAPCACHE)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Functions for defining drgn CLI commands and running them programmatically.

For documentation of available commands and command syntax, see
:doc:`commands`.
"""

from __future__ import annotations

import argparse
import collections
import contextlib
import dataclasses
import functools
import re
import subprocess
import sys
import textwrap
import traceback
import types
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Literal,
    NamedTuple,
    NoReturn,
    Optional,
    Protocol,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    overload,
)

from drgn.helpers.common.format import escape_ascii_string

if TYPE_CHECKING:
    if sys.version_info < (3, 11):
        from typing_extensions import Self, assert_never
    else:
        from typing import Self, assert_never  # novermin
    from _typeshed import SupportsWrite

import _drgn_util.argparseformatter
from _drgn_util.typingutils import copy_method_params
from drgn import Program, ProgramFlags

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)
T_contra = TypeVar("T_contra", contravariant=True)


_WORD_CHARACTER = r"""[^\s"#&'();<>\\`|]"""
_WORD_CHARACTER_OR_HASH = _WORD_CHARACTER.replace("#", "")
_ESCAPED_OR_QUOTED_WORD = r"""\\.|"(?:\\.|[^"])*"|'[^']*'"""

_SHELL_TOKEN_REGEX = re.compile(
    "|".join(
        "(?P<%s>%s)" % pair
        for pair in (
            ("REDIRECT", "(?P<redirect_fd>[0-9]*)(?P<redirect_op><|>>|>)"),
            (
                "WORD",
                r"(?:" + _WORD_CHARACTER + r"|" + _ESCAPED_OR_QUOTED_WORD + r")(?:"
                # '#' in the middle of a word doesn't start a comment.
                + _WORD_CHARACTER_OR_HASH + r"|" + _ESCAPED_OR_QUOTED_WORD + r")*",
            ),
            # We pass everything after a '|' to the shell, so capture the rest.
            ("PIPELINE", r"\|.*"),
            ("WHITESPACE", r"\s+"),
            ("COMMENT", r"#.*"),
            ("MISMATCH", r"."),
        )
    )
)


@contextlib.contextmanager
def _redirect_and_pipe(
    redirections: Sequence[ShellRedirection], pipeline: Optional[str]
) -> Iterator[None]:
    saved = {
        "stdin": sys.stdin,
        "stdout": sys.stdout,
        "stderr": sys.stderr,
    }
    pipe_process = None
    exceptions = []
    try:
        if pipeline is not None:
            pipe_process = subprocess.Popen(
                pipeline,
                shell=True,
                stdin=subprocess.PIPE,
                stdout=sys.stdout,
                text=True,
            )
            sys.stdout = pipe_process.stdin

        for fd, op, path in redirections:
            if op == "<":
                if fd != 0:
                    raise NotImplementedError(
                        f"redirecting input fd {fd} is not supported"
                    )
                attr = "stdin"
                mode = "r"
            elif op == ">" or op == ">>":
                if fd == 1:
                    attr = "stdout"
                elif fd == 2:
                    attr = "stderr"
                else:
                    raise NotImplementedError(
                        f"redirecting output fd {fd} is not supported"
                    )
                if op == ">":
                    mode = "w"
                elif op == ">>":
                    mode = "a"
                else:
                    assert_never(op)
            else:
                assert_never(op)
            if getattr(sys, attr) != saved[attr]:
                try:
                    getattr(sys, attr).close()
                except BrokenPipeError:
                    pass
            setattr(sys, attr, open(path, mode))

        try:
            yield
        except BrokenPipeError:
            pass
    except Exception as e:
        exceptions.append(e)
    finally:
        for attr, file in saved.items():
            if getattr(sys, attr) != file:
                try:
                    getattr(sys, attr).close()
                except BrokenPipeError:
                    pass
                except Exception as e:
                    exceptions.append(e)
                setattr(sys, attr, file)

        if pipe_process is not None:
            try:
                if pipe_process.wait() != 0:
                    raise CommandExitStatusError(
                        pipe_process.args,  # type: ignore[arg-type]
                        pipe_process.returncode,
                    )
            except Exception as e:
                exceptions.append(e)

        # Since Python 3.11, we can raise an exception group with multiple
        # exceptions. Otherwise, just raise the first one.
        if sys.version_info >= (3, 11) and len(exceptions) > 1:
            raise ExceptionGroup("errors while running command", exceptions)  # novermin
        elif exceptions:
            raise exceptions[0]


def _write_command_error(
    writer: SupportsWrite[str], e: Exception, *, prefix: str = "drgn"
) -> None:
    if isinstance(
        e, (CommandArgumentError, CommandExitStatusError, CommandNotFoundError)
    ):
        # These errors are already formatted, so don't add the "error:" prefix.
        writer.write(f"{prefix}: {e}\n")
    elif isinstance(e, SyntaxError):
        writer.write(f"{prefix}: syntax error: {e}\n")
    else:
        writer.write(f"{prefix}: error: {e}\n")


def run_command(
    prog: Program,
    command: str,
    *,
    globals: Optional[Dict[str, Any]] = None,
    onerror: Optional[Callable[[Exception], None]] = None,
) -> Any:
    """
    Run a drgn command.

    :param prog: Program to run command on.
    :param command: Command string (including arguments, redirections, pipes,
        etc.) to run.
    :param globals: Global variables to run command with. Defaults to the
        caller's global variables.
    :param onerror: Callback for handling errors (e.g., logging them). By
        default, errors are raised as exceptions.
    :return: The command's return value. Note that most commands return
        ``None``.
    """
    if globals is None:
        globals = sys._getframe(1).f_globals
    return DEFAULT_COMMAND_NAMESPACE.run(
        prog, command, globals=globals, onerror=onerror
    )


class CommandError(Exception):
    """Error raised when a command fails."""

    pass


class CommandArgumentError(CommandError):
    """Error raised when a command's arguments cannot be parsed."""

    pass


class CommandExitStatusError(CommandError):
    """Error raised when a command pipeline exits with a non-zero status."""

    command: str
    """Command pipeline that failed."""

    exit_status: int
    """Exit status, or negative signal number if terminated by a signal."""

    def __init__(self, command: str, exit_status: int) -> None:
        self.command = command
        self.exit_status = exit_status

    def __str__(self) -> str:
        return f"{self.command!r} exited with status {self.exit_status}"


class CommandNotFoundError(CommandError):
    """Error raised when a command cannot be found."""

    name: str
    """Command name that could not be found."""

    def __init__(self, name: str) -> None:
        self.name = name

    def __str__(self) -> str:
        return f"{self.name}: command not found"


class CommandNamespace:
    """
    Command namespace.

    Commands are registered to a specific namespace and can only be found or
    run in that namespace. This allows reusing drgn's command system for
    commands that shouldn't be exposed directly to the CLI.
    """

    def __init__(
        self,
        *,
        func_name_prefix: str = "_cmd_",
        argparse_types: Sequence[Tuple[str, Callable[[str], Any]]] = (),
    ) -> None:
        """
        Create a command namespace.

        :param func_name_prefix: Function name prefix expected by
            :func:`command()`, :func:`custom_command()`, and
            :func:`raw_command()`.
        :param argparse_types: ``(name, callable)`` tuples to register as
            argparse types for commands in this namespace. See
            :meth:`argparse.ArgumentParser.register()`.
        """

        self._commands: Dict[str, Command[object]] = {}
        self._func_name_prefix = func_name_prefix
        self._argparse_types = argparse_types

    def register(self, name: str, command: Command[Any]) -> None:
        """Register a command in this namespace."""
        self._commands[name] = command

    def lookup(self, prog: Program, name: str) -> Command[object]:
        """
        Find a command in this namespace.

        :param name: Command name.
        :raises CommandNotFoundError: if a command with the given name is not found
        """
        command = self._commands.get(name)
        if command is None or not command.enabled(prog):
            raise CommandNotFoundError(name)
        return command

    def enabled(self, prog: Program) -> Iterable[Tuple[str, Command[object]]]:
        """
        Get all enabled commands in this namespace.

        :return: Iterable of ``(name, command)`` tuples.
        """
        return (
            (name, command)
            for name, command in self._commands.items()
            if command.enabled(prog)
        )

    def run(
        self,
        prog: Program,
        command: str,
        *,
        globals: Optional[Dict[str, Any]] = None,
        onerror: Optional[Callable[[Exception], None]] = None,
    ) -> Any:
        """
        Run a drgn command in this namespace.

        Arguments and return value are the same as :func:`run_command()`.
        """
        if globals is None:
            globals = sys._getframe(1).f_globals

        try:
            kwargs = {"globals": globals}
            name, tail, command_obj, pager = self._resolve(prog, command, kwargs)
            parsed = command_obj.parse(tail)

            pipeline = parsed.pipeline
            if (
                pipeline is None
                and not any(redirection.fd == 1 for redirection in parsed.redirections)
                and pager is not None
            ):
                # We can only pipe to a pager if stdout is a file descriptor.
                try:
                    sys.stdout.fileno()
                except (AttributeError, OSError):
                    pass
                else:
                    pipeline = pager

            with _redirect_and_pipe(parsed.redirections, pipeline):
                return command_obj.run(prog, name, parsed.args, **kwargs)
        except Exception as e:
            if onerror is None:
                raise
            if sys.version_info >= (3, 11) and isinstance(
                e, ExceptionGroup
            ):  # novermin
                for e2 in e.exceptions:
                    onerror(e2)
            else:
                onerror(e)

    def _resolve(
        self, prog: Program, command: str, kwargs: Dict[str, Any]
    ) -> Tuple[str, str, Command[object], Optional[str]]:
        command = command.lstrip()
        match = _SHELL_TOKEN_REGEX.match(command)
        if not match or match.lastgroup != "WORD":
            raise SyntaxError("expected command name")

        name = unquote_shell_word(match.group())
        return name, command[match.end() :].lstrip(), self.lookup(prog, name), None


DEFAULT_COMMAND_NAMESPACE: CommandNamespace = CommandNamespace()
"""Default command namespace used by the drgn CLI."""


class Command(Protocol[T]):
    """
    Command implementation.

    Commands can usually be defined with :func:`command()`,
    :func:`custom_command()`, or :func:`raw_command()`, but this can be used
    when even more control is needed.
    """

    def description(self) -> str:
        """Return a one-line description of the command."""
        ...

    def format_usage(self) -> str:
        """Return the usage string."""
        ...

    def format_help(self, *, indent: str = "") -> str:
        """
        Return the help string.

        :param indent: String to prepend to each line.
        """
        ...

    def enabled(self, prog: Program) -> bool:
        """
        Return whether the command should be enabled for the given program.
        """
        ...

    def parse(self, source: str, /) -> ParsedCommand[T]:
        """
        Parse the command string after the command name.

        This must separate redirections and pipes from arguments.

        :param source: Command string after command name (arguments,
            redirections, pipes, etc.).
        """
        ...

    def run(
        self,
        prog: Program,
        name: str,
        args: T,
        /,
        *,
        globals: Dict[str, Any],
    ) -> Any:
        """
        Run the command.

        :param prog: Program.
        :param name: Name that the command was invoked as.
        :param args: Command arguments returned in :attr:`ParsedCommand.args`
            by :meth:`parse()`.

        Additional keyword arguments may also be passed. To allow for future
        extensions, implementations of this method should add a variable
        keyword parameter (i.e., ``**kwargs``) and ignore unneeded keyword
        arguments:

        .. code-block:: python3

            class MyCommand:
                ...
                def run(
                    self, prog: Program, name: str, args: Sequence[str], **kwargs: Any
                ) -> None:
                    ...

        The current keyword parameters are:

        :param globals: Dictionary of global variables.

        :return: Anything, but usually ``None``.
        """
        ...


@dataclasses.dataclass(frozen=True)
class ParsedCommand(Generic[T]):
    """Parsed command string."""

    args: T
    """Arguments."""

    redirections: Sequence[ShellRedirection] = ()
    """Shell redirections."""

    pipeline: Optional[str] = None
    """Pipeline."""


class ShellRedirection(NamedTuple):
    """Shell redirection."""

    fd: int
    """File descriptor to redirect."""

    op: RedirectOp
    """Redirection operator (``"<"``, ``">"``, etc.)."""

    path: str
    """Path to redirect to."""


RedirectOp = Literal["<", ">", ">>"]


def parse_shell_command(
    source: str, unquote: bool = True
) -> ParsedCommand[Sequence[str]]:
    """
    Parse a shell command string.

    :param source: Command string after command name (arguments,
        redirections, pipes, etc.).
    :param unquote: Whether to unquote/unescape arguments.
    """
    args: List[str] = []
    redirections: List[ShellRedirection] = []
    pipeline = None
    redirection = None

    for match in _SHELL_TOKEN_REGEX.finditer(source):
        kind = match.lastgroup
        value = match.group()
        if kind == "WORD":
            if redirection is None:
                if unquote:
                    value = unquote_shell_word(value)
                args.append(value)
            else:
                redirections.append(
                    ShellRedirection(
                        fd=redirection[0],
                        op=redirection[1],
                        path=unquote_shell_word(value),
                    )
                )
                redirection = None
        elif kind == "PIPELINE":
            # NB: since this is called with the substring after the command
            # name, unlike sh, we allow a string starting with '|'.
            if redirection is not None:
                raise SyntaxError("unexpected '|'")
            pipeline = value[1:].strip()
            if not pipeline:
                raise SyntaxError("unexpected end of input")
        elif kind == "REDIRECT":
            if redirection is not None:
                raise SyntaxError(f"unexpected {value!r}")
            redirect_op: RedirectOp = match.group("redirect_op")  # type: ignore[assignment]
            if match.group("redirect_fd"):
                redirect_fd = int(match.group("redirect_fd"))
            elif redirect_op == "<":
                redirect_fd = 0
            elif redirect_op == ">" or redirect_op == ">>":
                redirect_fd = 1
            else:
                assert_never(redirect_op)
            redirection = (redirect_fd, redirect_op)
        elif kind in ("WHITESPACE", "COMMENT"):
            pass
        else:
            raise SyntaxError(f"unexpected {value!r}")
    if redirection is not None:
        raise SyntaxError("unexpected end of input")
    return ParsedCommand(args, redirections, pipeline)


def _unquote_repl(match: "re.Match[str]") -> str:
    s = match.group()
    if s[0] == "\\":
        return s[1]
    elif s[0] == '"':
        return re.sub(r'\\([$`"\\\n])', r"\1", s[1:-1])
    else:
        assert s[0] == "'"
        return s[1:-1]


def unquote_shell_word(word: str) -> str:
    r"""
    Unquote/unescape a quoted and/or escaped word in shell syntax.

    >>> print(unquote_shell_word(r"f'o'o\ bar"))
    foo bar
    """
    return re.sub(_ESCAPED_OR_QUOTED_WORD, _unquote_repl, word)


@overload
def _sanitize_rst(s: str) -> str: ...


@overload
def _sanitize_rst(s: None) -> None: ...


def _sanitize_rst(s: Optional[str]) -> Optional[str]:
    if not s:
        return s
    return re.sub(r"\\(.)|[*\\]", r"\1", s)


class _DrgnCommandArgumentParser(argparse.ArgumentParser):
    def exit(self, status: int = 0, message: Optional[str] = None) -> NoReturn:
        if message is None:
            raise CommandArgumentError()
        else:
            raise CommandArgumentError(message.rstrip("\n"))


def _add_argument(
    parser: Any, arg: Union[argument, argument_group, mutually_exclusive_group]
) -> None:
    if isinstance(arg, argument):
        kwargs = arg.kwargs
        # argparse needs % to be escaped in help strings.
        if "%" in (help := kwargs.get("help", "")):
            kwargs = dict(kwargs, help=help.replace("%", "%%"))
        parser.add_argument(*arg.args, **kwargs)
    elif isinstance(arg, argument_group):
        group = parser.add_argument_group(title=arg.title, description=arg.description)
        for arg in arg.arguments:
            _add_argument(group, arg)
    elif isinstance(arg, mutually_exclusive_group):
        group = parser.add_mutually_exclusive_group(required=arg.required)
        for arg in arg.arguments:
            _add_argument(group, arg)
    else:
        assert_never(arg)


def _command_name(
    name: Optional[str],
    func: Callable[..., Any],
    prefix: str,
) -> str:
    if name is None:
        if not hasattr(func, "__name__"):
            raise ValueError("callable doesn't have __name__; pass name explicitly")
        match = re.fullmatch(prefix + r"(\w+)", func.__name__)
        if not match:
            raise ValueError(
                f"{func.__name__!r} doesn't start with {prefix}; "
                "rename it or pass name explicitly"
            )
        name = match.group(1)

    return name


def _decimal_or_hexadecimal(s: str) -> int:
    try:
        return int(s)
    except ValueError:
        return int(s, 16)


def _create_parser(
    *,
    name: str,
    usage: Optional[str] = None,
    description: Optional[str] = None,
    epilog: Optional[str] = None,
    arguments: Sequence[Union[argument, argument_group, mutually_exclusive_group]],
    types: Sequence[Tuple[str, Callable[[str], Any]]],
) -> argparse.ArgumentParser:
    parser = _DrgnCommandArgumentParser(
        prog=name,
        description=description,
        usage=_sanitize_rst(usage),
        epilog=epilog,
        formatter_class=_drgn_util.argparseformatter.MultilineHelpFormatter,
        add_help=False,
        allow_abbrev=False,
    )
    parser.register("type", "hexadecimal", lambda s: int(s, 16))
    parser.register("type", "decimal_or_hexadecimal", _decimal_or_hexadecimal)
    for type_name, type_func in types:
        parser.register("type", type_name, type_func)
    for arg in arguments:
        _add_argument(parser, arg)
    return parser


def _always_enabled(prog: Program) -> bool:
    return True


def command(
    *,
    name: Optional[str] = None,
    description: str,
    usage: Optional[str] = None,
    long_description: Optional[str] = None,
    epilog: Optional[str] = None,
    arguments: Sequence[Union[argument, argument_group, mutually_exclusive_group]] = (),
    enabled: Callable[[Program], bool] = _always_enabled,
    namespace: CommandNamespace = DEFAULT_COMMAND_NAMESPACE,
) -> CommandFuncDecorator:
    """
    Decorator to register a command.

    Commands registered with this decorator parse options specified in
    *arguments* using :mod:`argparse`. See :func:`custom_command()` and
    :func:`raw_command()` for alternatives that doesn't use :mod:`argparse`.

    See :class:`CommandFunc` for the signature of the command function.

    Descriptions and help strings may have multiple lines (unlike the default
    :mod:`argparse` formatting).

    There are also a couple of argparse argument types provided for convenience:

    1. ``"hexadecimal"``: a hexadecimal integer.
    2. ``"decimal_or_hexadecimal``: a decimal or hexadecimal integer,
       preferring decimal if ambiguous.

    .. code-block:: python3

        import argparse
        from typing import Any

        from drgn import Program
        from drgn.commands import command


        @command(description="do nothing")
        def _cmd_true(
            prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
        ) -> None:
            pass

    :param name: Command name. If not given, the name of the decorated function
        must begin with ``_cmd_``, and the command name is the function name
        with that prefix removed.
    :param description: Mandatory one-line description of the command.
    :param usage: Usage string in reStructuredText format. Generated from
        *arguments* automatically if not given.
    :param long_description: Optional longer description of the command.
    :param epilog: Optional additional information to show at the end of help
        output.
    :param arguments: Arguments, argument groups, and mutually exclusive groups
        accepted by the command.
    :param enabled: Callback returning whether the command should be enabled
        for a given program. Defaults to always enabled.
    :param namespace: Namespace to register command to.
    """

    def decorator(func: CommandFunc) -> CommandFunc:
        command_name = _command_name(name, func, namespace._func_name_prefix)

        parser = _create_parser(
            name=command_name,
            usage=usage,
            description=long_description,
            epilog=epilog,
            arguments=arguments,
            types=namespace._argparse_types,
        )

        namespace.register(
            command_name,
            _ArgparseCommand(
                func=func, parser=parser, description=description, enabled=enabled
            ),
        )

        return func

    return decorator


class CommandFunc(Protocol):
    """Signature of a command function for :func:`command()`."""

    def __call__(
        self,
        prog: Program,
        name: str,
        args: argparse.Namespace,
        /,
        *,
        parser: argparse.ArgumentParser,
        globals: Dict[str, Any],
    ) -> Any:
        """
        :param prog: Program.
        :param name: Name that the command was invoked as.
        :param args: Parsed arguments as an :class:`argparse.Namespace`.
        :param parser: Argument parser.

        See :meth:`Command.run()` for other keyword parameters.

        :return: Anything, but usually ``None``.
        """
        ...


CommandFuncDecorator = Callable[[CommandFunc], CommandFunc]


class argument:
    @copy_method_params(argparse.ArgumentParser.add_argument)
    def __init__(self, /, *args: Any, **kwargs: Any) -> None:
        """
        Command line argument for :func:`command()`.

        Arguments are the same as :meth:`argparse.ArgumentParser.add_argument()`.

        .. code-block:: python3

            import argparse
            from typing import Any

            from drgn import Program
            from drgn.commands import argument, command


            @command(
                description="concatenate and print files",
                arguments=(
                    argument("-n", help="number output lines"),
                    argument("file", nargs="*", help="input files"),
                ),
            )
            def _cmd_cat(
                prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
            ) -> None:
                ...
        """
        self.args = args
        self.kwargs = kwargs


drgn_argument: argument = argument(
    "--drgn",
    action="store_true",
    help="print how to do the equivalent operation using drgn APIs",
)
"""
Command line argument for :func:`command()` named ``--drgn`` indicating that
the command should print how to do the equivalent operation using drgn APIs.

This helps onboard users from simple commands to more complex use cases
requiring programmability.

By convention, it should normally be the last argument.

.. code-block:: python3

    import argparse
    from typing import Any

    from drgn import Program
    from drgn.commands import argument, command, drgn_argument


    @command(
        description="print a stack trace"
        arguments=(
            argument("tid", type=int, help="thread ID"),
            drgn_argument,
        ),
    )
    def _cmd_bt(
        prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
    ) -> None:
        if args.drgn:
            print("trace = stack_trace(tid)")
            print("print(trace)")
            return
        print(prog.stack_trace(args.tid))
"""


class argument_group:
    def __init__(
        self,
        /,
        *arguments: Union[argument, mutually_exclusive_group],
        title: str,
        description: Optional[str] = None,
    ) -> None:
        """
        Group of arguments for :func:`command()`.

        See :meth:`argparse.ArgumentParser.add_argument_group()`.

        .. code-block:: python3

            import argparse
            from typing import Any

            from drgn import Program
            from drgn.commands import argument, argument_group, command


            @command(
                description="print matching lines",
                arguments=(
                    argument_group(
                        argument("-A", type=int, help="print lines after match"),
                        argument("-B", type=int, help="print lines before match"),
                        title="context line options",
                    )
                    argument("pattern", help="pattern to search for"),
                ),
            )
            def _cmd_grep(
                prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
            ) -> None:
                ...

        :param arguments: Arguments and mutually exclusive groups in the group.
        :param title: Title to group arguments under.
        :param description: Description of group.
        """
        self.arguments = arguments
        self.title = title
        self.description = description


class mutually_exclusive_group:
    def __init__(self, /, *arguments: argument, required: bool = False) -> None:
        """
        Mutually exclusive group of arguments for :func:`command()`.

        See :meth:`argparse.ArgumentParser.add_mutually_exclusive_group()`.

        This is often nested in an :func:`argument_group()`:

        .. code-block:: python3

            import argparse
            from typing import Any

            from drgn import Program
            from drgn.commands import (
                argument,
                argument_group,
                command,
                mutually_exclusive_group,
            )


            @command(
                description="print matching lines",
                arguments=(
                    argument_group(
                        mutually_exclusive_group(
                            argument(
                                "-E",
                                action="store_true",
                                help="use extended regular expressions",
                            ),
                            argument(
                                "-F",
                                action="store_true",
                                help="match a fixed string",
                            ),
                        ),
                        title="pattern syntax",
                    )
                    argument("pattern", help="pattern to search for"),
                ),
            )
            def _cmd_grep(
                prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
            ) -> None:
                ...

        It can also be passed directly to :func:`command()`:

        .. code-block:: python3

            @command(
                description="print matching lines",
                arguments=(
                    mutually_exclusive_group(
                        argument(
                            "-E",
                            action="store_true",
                            help="use extended regular expressions",
                        ),
                        argument(
                            "-F",
                            action="store_true",
                            help="match a fixed string",
                        ),
                    ),
                    argument("pattern", help="pattern to search for"),
                ),
            )
            def _cmd_grep(
                prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
            ) -> None:
                ...

        :param arguments: Arguments in the group.
        :param required: Whether exactly one of the arguments is required.
        """
        self.required = required
        self.arguments = arguments


def custom_command(
    *,
    name: Optional[str] = None,
    description: str,
    usage: Optional[str] = None,
    long_description: Optional[str] = None,
    epilog: Optional[str] = None,
    arguments: Optional[
        Sequence[Union[argument, argument_group, mutually_exclusive_group]]
    ] = None,
    parse: Callable[[str], ParsedCommand[T]],
    enabled: Callable[[Program], bool] = _always_enabled,
    namespace: CommandNamespace = DEFAULT_COMMAND_NAMESPACE,
) -> CustomCommandFuncDecorator[T]:
    """
    Decorator to register a command with its own custom syntax (instead of
    command line options).

    See :class:`CustomCommandFunc` for the signature of the command function.

    .. code-block:: python3

        import ast
        import dataclasses
        import re
        from typing import Any

        from drgn import Program
        from drgn.commands import ParsedCommand, custom_command, parse_shell_command


        def _parse_literal_eval(source: str) -> ParsedCommand[str]:
            match = re.fullmatch(r"([^<>|]*)(.*)", source)
            parsed = parse_shell_command(match.group(2))
            if parsed.args:
                raise SyntaxError("arguments after redirections are not supported")
            return dataclasses.replace(parsed, args=match.group(1))


        @custom_command(
            description="evaluate a Python literal",
            usage="**literal_eval** *EXPR*",
            long_description="This evaluates and returns a Python literal"
            " (for example, a string, integer, list, etc.).",
            parse=_parse_literal_eval,
        )
        def _cmd_literal_eval(
            prog: Program, name: str, args: str, **kwargs: Any
        ) -> Any:
            return ast.literal_eval(args)

    :param name: Command name. If not given, the name of the decorated function
        must begin with ``_cmd_``, and the command name is the function name
        with that prefix removed.
    :param description: Mandatory one-line description of the command.
    :param usage: Usage string in reStructuredText format. Mandatory if
        *arguments* not given, otherwise generated from *arguments*
        automatically if not given.
    :param long_description: Longer description of the command. Mandatory if
        *arguments* not given, optional otherwise.
    :param epilog: Optional additional information to show at the end of help
        output.
    :param arguments: Arguments, argument groups, and mutually exclusive groups
        accepted by the command. Used **only** for generating help output.
    :param parse: Callback to parse the command string after the command name.
        This must separate redirections and pipes from arguments.
    :param enabled: Callback returning whether the command should be enabled
        for a given program. Defaults to always enabled.
    :param namespace: Namespace to register command to.
    """

    if arguments is None:
        if usage is None:
            raise TypeError("usage is mandatory if arguments not given")
        if long_description is None:
            raise TypeError("long_description is mandatory if arguments not given")
        arguments = ()

    def decorator(func: CustomCommandFunc[Any]) -> CustomCommandFunc[Any]:
        command_name = _command_name(name, func, namespace._func_name_prefix)

        parser = _create_parser(
            name=command_name,
            usage=usage,
            description=long_description,
            epilog=epilog,
            arguments=arguments,
            types=namespace._argparse_types,
        )

        namespace.register(
            command_name,
            _CustomCommand(
                func=func,
                parser=parser,
                description=description,
                parse=parse,
                enabled=enabled,
            ),
        )

        return func

    return decorator


class CustomCommandFunc(Protocol[T_contra]):
    """Signature of a custom command function for :func:`custom_command()`."""

    def __call__(
        self,
        prog: Program,
        name: str,
        args: T_contra,
        /,
        *,
        parser: argparse.ArgumentParser,
        globals: Dict[str, Any],
    ) -> Any:
        """
        :param prog: Program.
        :param name: Name that the command was invoked as.
        :param args: Command arguments parsed by *parse* argument to
            :func:`custom_command()`.
        :param parser: Argument parser.

        See :meth:`Command.run()` for other keyword parameters.

        :return: Anything, but usually ``None``.
        """
        ...


CustomCommandFuncDecorator = Callable[[CustomCommandFunc[T]], CustomCommandFunc[T]]


raw_command = functools.partial(custom_command, parse=ParsedCommand)
"""
Special case of :class:`custom_command()` where the arguments are the command
string verbatim.

Note that this cannot support redirections or pipes.

.. code-block:: python3

    from typing import Any

    from drgn import Program
    from drgn.commands import argument, raw_command


    @raw_command(
        description="display text verbatim",
        arguments=(
            argument("text", help="text to display"),
        ),
    )
    def _cmd_echo(
        prog: Program, name: str, args: str, **kwargs: Any
    ) -> Any:
        print(args)
"""


class _CustomCommand(Generic[T]):
    def __init__(
        self,
        func: CustomCommandFunc[T],
        parser: argparse.ArgumentParser,
        description: str,
        parse: Callable[[str], ParsedCommand[T]],
        enabled: Callable[[Program], bool],
    ) -> None:
        self._func = func
        self._parser = parser
        self._description = description
        self.parse = parse
        self.enabled = enabled

    def description(self) -> str:
        return self._description

    def format_usage(self) -> str:
        usage = self._parser.format_usage()
        if usage.startswith("usage: "):
            usage = usage[len("usage: ") :]
        return usage.rstrip()

    def format_help(self, *, indent: str = "") -> str:
        usage = self._parser.format_usage()
        help = self._parser.format_help()
        if help.startswith(usage):
            help = help[len(usage) :].lstrip()
        return textwrap.indent(help, indent).rstrip()

    def run(self, prog: Program, name: str, args: T, **kwargs: Any) -> Any:
        return self._func(prog, name, args, **kwargs, parser=self._parser)


class _ArgparseCommand(_CustomCommand[Sequence[str]]):
    def __init__(
        self,
        func: CommandFunc,
        parser: argparse.ArgumentParser,
        description: str,
        enabled: Callable[[Program], bool],
    ) -> None:
        super().__init__(
            self._run_argparse, parser, description, parse_shell_command, enabled
        )
        self._argparse_func = func

    def _run_argparse(
        self,
        prog: Program,
        name: str,
        args: Sequence[str],
        parser: argparse.ArgumentParser,
        **kwargs: Any,
    ) -> Any:
        return self._argparse_func(
            prog, name, parser.parse_args(args), **kwargs, parser=parser
        )


# Variant of repr() that prefers double-quoted strings (like Black).
def _repr_black(obj: Any) -> str:
    if not isinstance(obj, str):
        return repr(obj)
    r = repr(obj + "'")
    if r.endswith("'\""):
        return r[:-2] + '"'
    else:
        assert r.endswith(r"\''")
        return r[:-3] + "'"


# repr() a bytes-like object as a raw bytes literal (if possible), preferring
# double quotes.
def _repr_raw_bytes(buffer: Sequence[int]) -> str:
    contains_single_quote = ord("'") in buffer
    contains_double_quote = ord('"') in buffer

    # If the string contains both double and single quotes, or if it ends with
    # a backslash, then it can't be a raw literal.
    if (contains_double_quote and contains_single_quote) or (
        buffer and buffer[-1] == ord("\\")
    ):
        prefix = "b"
        escape_backslash = True
    else:
        prefix = "rb"
        escape_backslash = False

    quote = '"' if not contains_double_quote or contains_single_quote else "'"

    escaped = escape_ascii_string(
        buffer,
        escape_single_quote=quote == "'",
        escape_double_quote=quote == '"',
        escape_backslash=escape_backslash,
    )

    return f"{prefix}{quote}{escaped}{quote}"


class DrgnCodeBuilder:
    """
    Helper class for generating code for :func:`drgn_argument`.

    This handles joining multiple fragments of code and adding imports. Imports
    are deduplicated, sorted, and prepended to the final output.
    """

    def __init__(self, prog: Program) -> None:
        self._prog = prog
        self._code: List[str] = []
        self._imports: Dict[str, Set[str]] = collections.defaultdict(set)
        self._blocks: List[DrgnCodeBlockContext] = []

    def append(self, code: str) -> None:
        """Append a code fragment to the output."""
        if not code:
            return

        if not self._blocks:
            self._code.append(code)
            return

        indent = self._blocks[-1]._indent
        should_indent = not self._code or self._code[-1].endswith("\n")
        for line in code.splitlines(keepends=True):
            if should_indent and line != "\n":
                self._code.append(indent)

            self._code.append(line)
            should_indent = True

    def begin_block(self, code: str, end: str = "") -> DrgnCodeBlockContext:
        """
        Append a code fragment that begins an indented block.

        This can be used to wrap code that may or may not need to be executed
        in a conditional or loop.

        Subsequent lines added with :meth:`append()` will be indented
        accordingly until the block is ended with :meth:`end_block()`:

        .. code-block:: python3

            if condition:
                code.begin_block(f"if {condition}:\\n")
            code.append("do_something()\\n")
            if condition:
                code.end_block()

        Alternatively, this can be used as a context manager:

        .. code-block:: python3

            with code.begin_block(f"while {condition}:\\n")
                code.append("do_something()\\n")

        :param code: Code fragment whose last non-whitespace line begins the
            block (e.g., an ``if`` or ``for`` statement). This can also be
            empty to begin a "fake" block that doesn't affect indentation but
            still pairs with a :meth:`end_block()` call.
        :param end: Code fragment to append when the block is ended. This is
            useful for ``try`` statements.
        """
        if code:
            match = re.fullmatch(
                r"""
                # Skip everything until we find...
                (?s:.)*
                # ... the last line that is not only whitespace, and capture
                # its indentation.
                ^([ \t]*)\S.*\n
                # Ignore zero or more trailing lines of only whitespace.
                (?:\s*\n)?
                """,
                code,
                flags=re.MULTILINE | re.VERBOSE,
            )
            if not match:
                if code.endswith("\n"):
                    raise ValueError("code is all whitespace")
                else:
                    raise ValueError("code does not end with newline")
            indent = match.group(1) + "    "

            self.append(code)
        else:
            indent = ""

        if self._blocks:
            indent = self._blocks[-1]._indent + indent
        block = DrgnCodeBlockContext(self, indent, end)
        self._blocks.append(block)
        return block

    def end_block(self) -> None:
        """
        End the most recently begun indented block.

        See :meth:`begin_block()`.
        """
        self._blocks[-1].end()

    def add_import(self, module: str) -> None:
        """
        Add an import to the output.

        :param modules: Module to import.
        """
        self._imports[module].add("")

    def add_from_import(self, module: str, *names: str) -> None:
        """
        Add a ``from module import name1, name2, ...`` import to the output.

        :param module: Module name.
        :param names: Names to import from *module*.
        """
        self._imports[module].update(names)

    def begin_retry_loop_if_live(self, num_attempts: int) -> "DrgnCodeBlockContext":
        """
        Begin a loop that retries on transient :class:`~drgn.FaultError` or
        :class:`~drgn.helpers.ValidationError` errors if the program is live.

        This must be paired with :meth:`end_block()` or used as a context
        manager.

        :param num_attempts: Maximum number of attempts.
        """
        if not (self._prog.flags & ProgramFlags.IS_LIVE):
            return self.begin_block("")

        self.add_from_import("drgn", "FaultError")
        self.add_from_import("drgn.helpers", "ValidationError")
        return self.begin_block(
            f"""\
# This is racy. Retry a limited number of times.
for attempts_remaining in range({num_attempts}, 0, -1):
    try:
""",
            """\
        break
    except (FaultError, ValidationError):
        if attempts_remaining == 1:
            raise
""",
        )

    def get(self) -> str:
        """Get the output as a string."""
        assert not self._blocks

        parts: List[str] = []
        first_party_imports: List[str] = []
        for module, names in sorted(self._imports.items()):
            if module == "drgn" or module.startswith("drgn."):
                target = first_party_imports
            else:
                target = parts

            if "" in names:
                names.remove("")
                target.append(f"import {module}\n")

            if names:
                sorted_names = sorted(names)
                line = f"from {module} import {', '.join(sorted_names)}\n"
                # 88 (the default Black line length) + 1 for the newline.
                if len(line) <= 89:
                    target.append(line)
                else:
                    target.append(f"from {module} import (\n")
                    for name in sorted_names:
                        target.append(f"    {name},\n")
                    target.append(")\n")

        if parts and first_party_imports:
            parts.append("\n")
        parts.extend(first_party_imports)

        if parts and self._code:
            parts.append("\n\n")
        parts.extend(self._code)

        return "".join(parts)

    def print(self) -> None:
        """Write the output to standard output."""
        sys.stdout.write(self.get())

    def has_declaration(self, variable: str) -> bool:
        """Return true if the variable is declared in the code."""
        for block in self._code:
            for line in block.split("\n"):
                if line.lstrip(" ").startswith(f"{variable} = "):
                    return True
        return False


class DrgnCodeBlockContext:
    def __init__(self, builder: DrgnCodeBuilder, indent: str, end: str) -> None:
        self._builder = builder
        self._indent = indent
        self._end = end

    def end(self) -> None:
        assert self._builder._blocks[-1] is self
        self._builder.append(self._builder._blocks.pop()._end)

    def __enter__(self) -> "Self":
        return self

    def __exit__(self, *exc_info: object) -> None:
        self.end()


def _parse_py_command(
    args: str,
) -> ParsedCommand[Union[types.CodeType, SyntaxError, None]]:
    for match in re.finditer(r"[|<>]", args):
        try:
            source = args[: match.start()]
            if not source or source.isspace():
                code = None
            else:
                code = compile(source, "<input>", "single")
        except SyntaxError:
            pass
        else:
            parsed = parse_shell_command(args[match.start() :])
            if parsed.args:
                # Don't allow extra arguments to be mixed in with redirections.
                raise SyntaxError("py does not support arguments after redirections")
            return dataclasses.replace(parsed, args=code)  # type: ignore[arg-type,return-value]
    else:
        # Fallback for no match: compile all the code as a "single" statement
        # so exec() still prints out the result. If there is a syntax error,
        # let the command handle it.
        if not args or args.isspace():
            return ParsedCommand(None)
        try:
            return ParsedCommand(compile(args, "<input>", "single"))
        except SyntaxError as e:
            return ParsedCommand(e)


# Print an exception without our own compile() frame, which could confuse the
# user.
def _print_py_command_exception(exc: BaseException) -> None:
    # Unfortunately, traceback objects are linked lists and there's no built-in
    # functionality to drop the last N frames of a traceback while printing.
    tb = exc.__traceback__
    count = 0
    while tb:
        count += 1
        tb = tb.tb_next
    traceback.print_exception(type(exc), exc, exc.__traceback__, limit=1 - count)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
I/O Resources
-------------

The ``drgn.helpers.linux.ioport`` module provides helpers for working with I/O
resources.
"""

from typing import Iterator

from drgn import Object

__all__ = ("for_each_resource",)


def for_each_resource(root: Object) -> Iterator[Object]:
    """
    Iterate over all I/O resources starting from the given root resource.

    :param root: ``struct resource *``
    :return: Iterator of ``struct resource *`` objects.
    """
    yield root
    child = root.child.read_()
    while child:
        yield from for_each_resource(child)
        child = child.sibling.read_()



#!/usr/bin/env drgn
# Copyright (c) Canonical Ltd.
# SPDX-License-Identifier: LGPL-2.1-or-later

""" Script to dump slabinfo status using drgn"""

from typing import Iterator, Optional

from drgn import Object
from drgn.helpers.common.format import escape_ascii_string
from drgn.helpers.linux.list import list_for_each_entry, list_for_each_entry_reverse
from drgn.helpers.linux.slab import for_each_slab_cache

MAX_PARTIAL_TO_SCAN = 10000
OO_SHIFT = 16
OO_MASK = (1 << OO_SHIFT) - 1


def for_each_kmem_cache_node(slab_cache: Object) -> Iterator[Object]:
    """
    Iterate over all kmem_cache_node of specific slab cache.

    :return: Iterator of ``struct kmem_cache_node *`` objects.
    """
    for nid in range(0, prog["nr_node_ids"].value_()):
        yield slab_cache.node[nid]


def count_partial_free_approx(kmem_cache_node: Object) -> Optional[Object]:
    x = Object(prog, "unsigned long", 0)
    n = kmem_cache_node
    if n.nr_partial <= MAX_PARTIAL_TO_SCAN:
        for slab in list_for_each_entry(
            "struct slab", n.partial.address_of_(), "slab_list"
        ):
            x += slab.objects - slab.inuse
    else:
        scanned = 0
        for slab in list_for_each_entry(
            "struct slab", n.partial.address_of_(), "slab_list"
        ):
            x += slab.objects - slab.inuse
            scanned += 1
            if scanned == MAX_PARTIAL_TO_SCAN / 2:
                break
        for slab in list_for_each_entry_reverse(
            "struct slab", n.partial.address_of_(), "slab_list"
        ):
            x += slab.objects - slab.inuse
            scanned += 1
            if scanned == MAX_PARTIAL_TO_SCAN / 2:
                break
        x = x * n.nr_partial / scanned
        x = min(x, n.total_objects)
    return x


def oo_objects(kmem_cache_order_objects: Object) -> Optional[Object]:
    return kmem_cache_order_objects.x & OO_MASK


def oo_order(kmem_cache_order_objects: Object) -> Optional[Object]:
    return kmem_cache_order_objects.x >> OO_SHIFT


print(
    f"{'struct kmem_cache *':^20} | {'name':^20} | {'active_objs':^12} | {'num_objs':^12} | {'objsize':^8} | {'objperslab':^11} | {'pageperslab':^13}"
)
print(
    f"{'':-^20} | {'':-^20} | {'':-^12} | {'':-^12} | {'':-^8} | {'':-^11} | {'':-^13}"
)

for s in for_each_slab_cache(prog):
    nr_slabs = 0
    nr_objs = 0
    nr_free = 0
    for node in for_each_kmem_cache_node(s):
        nr_slabs += node.nr_slabs.counter.value_()
        nr_objs += node.total_objects.counter.value_()
        nr_free += count_partial_free_approx(node).value_()
    active_objs = nr_objs - nr_free
    num_objs = nr_objs
    active_slab = nr_slabs
    num_slabs = nr_slabs
    objects_per_slab = oo_objects(s.oo).value_()
    cache_order = oo_order(s.oo).value_()
    name = escape_ascii_string(s.name.string_(), escape_backslash=True)

    print(
        f"0x{s.value_():<18x} | {name:20.19s} | {active_objs:12} | {num_objs:12} | {s.size.value_():8} | {objects_per_slab:11} | {1<<cache_order:13}"
    )



# Copyright (c) Google LLC
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Stack Depot
-----------

The ``drgn.helpers.linux.stackdepot`` module provides helpers for working with
the stack trace storage from :linux:`include/linux/stackdepot.h` used by KASAN
and other kernel debugging tools.
"""

from typing import Optional

from drgn import Object, StackTrace, cast, reinterpret

__all__ = ("stack_depot_fetch",)


def stack_depot_fetch(handle: Object) -> Optional[StackTrace]:
    """
    Returns a stack trace for the given stack handle.

    :param handle: ``depot_stack_handle_t``
    :return: The stack trace, or ``None`` if not available.
    """
    prog = handle.prog_
    handle_parts = reinterpret("union handle_parts", handle)

    # Renamed in Linux kernel commit 961c949b012f ("lib/stackdepot: rename slab
    # to pool") (in v6.3).
    try:
        stack_pools = prog["stack_pools"]
    except KeyError:
        pool = prog["stack_slabs"][handle_parts.slabindex]
    else:
        # Linux kernel commit 3ee34eabac2a ("lib/stackdepot: fix first entry
        # having a 0-handle") (in v6.9-rc1) changed the meaning of pool_index.
        # Linux kernel commit a6c1d9cb9a68 ("stackdepot: rename pool_index to
        # pool_index_plus_1") (in v6.9-rc3) renamed pool_index to reflect the
        # new meaning. This will therefore be wrong for v6.9-rc[1-2] and
        # v6.8.[3-4].
        try:
            pool_index = handle_parts.pool_index_plus_1 - 1
        except AttributeError:
            pool_index = handle_parts.pool_index
        pool = stack_pools[pool_index]

    if not pool:
        return None

    # This has remained the same since the stack depot was introduced in Linux
    # kernel commit cd11016e5f52 ("mm, kasan: stackdepot implementation. Enable
    # stackdepot for SLAB") (in v4.6), when it was known as STACK_ALLOC_ALIGN.
    DEPOT_STACK_ALIGN = 4

    record = cast(
        "struct stack_record *", pool + (handle_parts.offset << DEPOT_STACK_ALIGN)
    )
    return prog.stack_trace_from_pcs([record.entries[x] for x in range(record.size)])



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Crash commands for iterating kernel collection types."""

import argparse
import sys
import typing
from typing import Any, Container, Literal, Optional, Sequence, Set, Tuple, Union

if typing.TYPE_CHECKING:
    if sys.version_info < (3, 11):
        from typing_extensions import assert_never
    else:
        from typing import assert_never  # novermin

from drgn import FaultError, Object, Program
from drgn.commands import (
    CommandArgumentError,
    DrgnCodeBuilder,
    _repr_black,
    argument,
    drgn_argument,
    mutually_exclusive_group,
)
from drgn.commands._crash.common import (
    CrashDrgnCodeBuilder,
    _addr_or_sym,
    _guess_type,
    _guess_type_name,
    _object_format_options,
    _parse_type_name_and_members,
    _parse_type_offset_arg,
    _prefer_object_lookup,
    _resolve_addr_or_sym,
    _resolve_type_offset_arg,
    _sanitize_member_name,
    crash_command,
)
from drgn.helpers.common.type import typeof_member
from drgn.helpers.linux.mapletree import mt_for_each
from drgn.helpers.linux.radixtree import radix_tree_for_each
from drgn.helpers.linux.rbtree import rbtree_inorder_for_each, rbtree_preorder_for_each
from drgn.helpers.linux.xarray import xa_for_each


def _append_get_members(
    code: DrgnCodeBuilder,
    name: str,
    members: Sequence[str],
    *,
    indent: str = "",
    avoid_names: Container[str] = (),
) -> None:
    if not members:
        code.append(indent + "...\n")
        return

    for member in members:
        sanitized_name = _sanitize_member_name(member)
        while sanitized_name == name or sanitized_name in avoid_names:
            sanitized_name += "_"

        if hasattr(Object, member):
            member = f"member_({_repr_black(member)})"

        code.append(f"{indent}{sanitized_name} = {name}.{member}\n")


def _append_list_head(
    code: DrgnCodeBuilder,
    name: str = "head",
    *,
    start: Union[str, int],
    head_node_offset_arg: Union[None, int, Tuple[str, str]],
    entry_name: Optional[str] = None,
) -> None:
    prog = code._prog

    if head_node_offset_arg is None:
        if isinstance(start, str) and _prefer_object_lookup(
            prog,
            "struct list_head" if entry_name is None else entry_name,
            start,
            strict_type_name=entry_name is None,
        ):
            code.append(f"{name} = prog[{_repr_black(start)}].address_of_()\n")
            return
    elif isinstance(head_node_offset_arg, tuple):
        if isinstance(start, str):
            if _prefer_object_lookup(
                prog, head_node_offset_arg[0], start, strict_type_name=False
            ):
                # Here we're trusting the user that the -O argument refers to a
                # list_head member.
                code.append(
                    f"{name} = prog[{_repr_black(start)}].{head_node_offset_arg[1]}.address_of_()\n"
                )
                return
        else:
            code.add_from_import("drgn", "Object")
            head_type_name = _guess_type_name(prog, head_node_offset_arg[0])
            # Same here, we're trusting the user that the -O argument refers to
            # a list_head member.
            code.append(
                f"{name} = Object(prog, {_repr_black(head_type_name)}, address={start:#x}).{head_node_offset_arg[1]}.address_of_()\n"
            )
            return

    if isinstance(start, str):
        code.append(f"address = prog.symbol({_repr_black(start)}).address\n")
        head_address = "address"
    else:
        head_address = f"{start:#x}"

    if head_node_offset_arg is None:
        head_offset = None
    elif isinstance(head_node_offset_arg, tuple):
        code.add_from_import("drgn", "offsetof")
        head_type_name = _guess_type_name(prog, head_node_offset_arg[0])
        head_offset = f"offsetof({_repr_black(head_type_name)}, {_repr_black(head_node_offset_arg[1])})"
    else:
        head_offset = str(head_node_offset_arg)
    if head_offset is not None:
        code.append(f"offset = {head_offset}\n")

    code.add_from_import("drgn", "Object")
    if entry_name is None:
        type_name = "struct list_head"
    else:
        type_name = _guess_type_name(prog, entry_name)
    code.append(f'{name} = Object(prog, "{type_name} *", {head_address}')
    if head_offset is not None:
        code.append(" + offset")
    code.append(")\n")


def _list_drgn_option_list_for_each(
    code: DrgnCodeBuilder,
    args: argparse.Namespace,
    *,
    start: Union[str, int],
    head_node_offset_arg: Union[None, int, Tuple[str, str]],
    node_member: Optional[str],
    entry_name: Optional[str],
    entry_members: Sequence[str],
) -> None:
    prog = code._prog

    _append_list_head(code, start=start, head_node_offset_arg=head_node_offset_arg)

    helper = "list_for_each"
    if entry_name is not None:
        helper += "_entry"
    if args.reverse:
        helper += "_reverse"

    code.add_from_import("drgn.helpers.linux.list", helper)

    if entry_name is None:
        code.append(
            f"""\
for node in {helper}(head):
    ...
"""
        )
    else:
        assert node_member is not None
        entry_type_name = _guess_type_name(prog, entry_name)
        code.append(
            f"""\
for entry in {helper}({_repr_black(entry_type_name)}, head, {_repr_black(node_member)}):
"""
        )
        _append_get_members(code, "entry", entry_members, indent="    ")

    code.print()


def _list_drgn_option_embedded_list_head(
    code: DrgnCodeBuilder,
    args: argparse.Namespace,
    *,
    start: Union[str, int],
    node_member: Optional[str],
    entry_name: Optional[str],
    entry_members: Sequence[str],
) -> None:
    _append_list_head(
        code,
        "first",
        start=start,
        head_node_offset_arg=None,
        entry_name=entry_name,
    )

    if entry_name is None:
        code.append(
            f"""\
node = first
while True:
    ...

    node = node.{"prev" if args.reverse else "next"}
    if node == first:
        break
"""
        )
    else:
        assert node_member is not None
        helper = f"list_{'prev' if args.reverse else 'next'}_entry"
        code.add_from_import("drgn.helpers.linux.list", helper)
        code.append(
            """\
entry = first
while True:
"""
        )

        _append_get_members(
            code, "entry", entry_members, indent="    ", avoid_names=("first",)
        )

        code.append(
            f"""\

    entry = {helper}(entry, {_repr_black(node_member)})
    if entry == first:
        break
"""
        )

    code.print()


def _list_drgn_option_next_member(
    code: DrgnCodeBuilder,
    args: argparse.Namespace,
    *,
    start: Union[str, int],
    node_member: str,
    entry_name: Optional[str],
    entry_members: Sequence[str],
) -> None:
    _append_list_head(
        code,
        "first",
        start=start,
        head_node_offset_arg=None,
        entry_name=entry_name,
    )

    code.append(
        """\
entry = first
while True:
"""
    )

    _append_get_members(
        code, "entry", entry_members, indent="    ", avoid_names=("first",)
    )

    code.append(
        f"""\

    next_entry = entry.{node_member}
    if not next_entry or next_entry == first or next_entry == entry"""
    )

    if args.end is not None:
        code.append(f" or next_entry.value_() == {args.end:#x}")

    code.append(
        """:
        break
    entry = next_entry
"""
    )

    code.print()


def _list_drgn_option(
    prog: Program,
    args: argparse.Namespace,
    *,
    next_offset_arg: Union[None, int, Tuple[str, str]],
    head_node_offset_arg: Union[None, int, Tuple[str, str]],
    node_offset_arg: Union[None, int, Tuple[str, str]],
    entry_name: Optional[str],
    entry_members: Sequence[str],
) -> None:
    # Note: we don't bother generating code for checking for duplicates or
    # stopping at the second entry.

    code = CrashDrgnCodeBuilder(prog)

    try:
        start: Union[str, int] = int(args.start, 16)
    except ValueError:
        start = args.start

    node_member = None
    if entry_name is None:
        if isinstance(next_offset_arg, tuple):
            # -o without -s prints the entry address but not the entry.
            entry_name, node_member = next_offset_arg
    else:
        # -o prints the entry addresses and -l prints the node addresses, but
        # we ignore that difference.
        if isinstance(next_offset_arg, tuple) and next_offset_arg[0] == entry_name:
            node_member = next_offset_arg[1]
        elif isinstance(node_offset_arg, tuple) and node_offset_arg[0] == entry_name:
            node_member = node_offset_arg[1]

    if args.list_head or head_node_offset_arg is not None:
        if (
            (
                # It makes sense to iterate over list_heads if either the
                # entry type+node member make sense...
                node_member is not None
                # ... or if there is no entry type.
                or (entry_name is None and next_offset_arg is None)
            )
            # Weird end values don't make sense for list_head.
            and args.end is None
        ):
            return _list_drgn_option_list_for_each(
                code,
                args,
                start=start,
                head_node_offset_arg=head_node_offset_arg,
                node_member=node_member,
                entry_name=entry_name,
                entry_members=entry_members,
            )
    elif args.embedded_list_head:
        if (
            node_member is not None or (entry_name is None and next_offset_arg is None)
        ) and args.end is None:
            return _list_drgn_option_embedded_list_head(
                code,
                args,
                start=start,
                node_member=node_member,
                entry_name=entry_name,
                entry_members=entry_members,
            )
    elif node_member is not None:
        return _list_drgn_option_next_member(
            code,
            args,
            start=start,
            node_member=node_member,
            entry_name=entry_name,
            entry_members=entry_members,
        )

    if isinstance(start, str):
        code.append(f"start = prog.symbol({_repr_black(start)}).address\n")
    else:
        code.append(f"start = {start:#x}")

    if entry_name is not None:
        entry_type_name = _guess_type_name(prog, entry_name)

    if isinstance(next_offset_arg, tuple):
        if next_offset_arg[0] == entry_name:
            next_type_name = entry_type_name
        else:
            next_type_name = _guess_type_name(prog, next_offset_arg[0])
        code.add_from_import("drgn", "offsetof")
        next_offset = f"offsetof({_repr_black(next_type_name)}, {_repr_black(next_offset_arg[1])})"
    elif isinstance(next_offset_arg, int):
        next_offset = str(next_offset_arg)
    else:
        next_offset = None

    next_or_prev = "prev" if args.reverse else "next"
    list_head_offset = None
    if args.list_head or args.embedded_list_head:
        list_head_offset = next_offset
        if args.reverse:
            code.add_from_import("drgn", "offsetof")
            next_offset = 'offsetof("struct list_head", "prev")'
        else:
            next_offset = None

    if list_head_offset is not None:
        code.append(f"list_head_offset = {list_head_offset}\n")
    if next_offset is not None:
        code.append(f"{next_or_prev}_offset = {next_offset}\n")

    entry_offset = None
    if node_offset_arg is not None:
        entry_offset = "entry_offset"
        if isinstance(node_offset_arg, tuple):
            if node_offset_arg[0] == entry_name:
                node_type_name = entry_type_name
            else:
                node_type_name = _guess_type_name(prog, node_offset_arg[0])
            code.add_from_import("drgn", "offsetof")
            code.append(
                f"entry_offset = offsetof({_repr_black(node_type_name)}, {_repr_black(node_offset_arg[1])})\n"
            )
        else:
            code.append(f"entry_offset = {node_offset_arg}\n")
    elif list_head_offset is not None:
        entry_offset = "list_head_offset"

    code.append("\n")

    if args.list_head or head_node_offset_arg is not None:
        if isinstance(head_node_offset_arg, tuple):
            if head_node_offset_arg[0] == entry_name:
                next_type_name = entry_type_name
            else:
                next_type_name = _guess_type_name(prog, head_node_offset_arg[0])
            code.add_from_import("drgn", "offsetof")
            code.append(
                f"start += offsetof({_repr_black(next_type_name)}, {_repr_black(head_node_offset_arg[1])})\n"
            )
        elif isinstance(head_node_offset_arg, int):
            code.append(f"start += {head_node_offset_arg}\n")

        code.append("ptr = prog.read_word(start")
        if next_offset is not None:
            code.append(f" + {next_or_prev}_offset")
        code.append(")\nwhile ptr != start:\n")
    else:
        if list_head_offset is not None:
            code.append("start += list_head_offset\n")
        code.append("ptr = start\nwhile True:\n")

    if entry_name:
        code.add_from_import("drgn", "Object")
        code.append(
            f'    entry = Object(prog, {_repr_black(entry_type_name + " *")}, ptr'
        )
        if entry_offset is not None:
            code.append(f" - {entry_offset}")
        code.append(")\n")
        if entry_members:
            _append_get_members(
                code,
                "entry",
                entry_members,
                indent="    ",
                avoid_names={
                    "start",
                    next_or_prev + "_offset",
                    "list_head_offset",
                    "entry_offset",
                },
            )
    elif list_head_offset is not None:
        code.append("    entry_ptr = ptr - list_head_offset\n")
    else:
        code.append("    ...\n")

    code.append(
        f"""\

    {next_or_prev}_ptr = prog.read_word(ptr"""
    )
    if next_offset is not None:
        code.append(f" + {next_or_prev}_offset")
    code.append(")\n    if")
    if not args.list_head and not args.embedded_list_head:
        code.append(f" not {next_or_prev}_ptr or")
    if not args.list_head and head_node_offset_arg is None:
        code.append(f" {next_or_prev}_ptr == start or")
    code.append(f" {next_or_prev}_ptr == ptr")

    if args.end is not None:
        code.append(f" or {next_or_prev}_ptr == {args.end:#x}")

    code.append(
        f""":
        break
    ptr = {next_or_prev}_ptr
"""
    )

    code.print()


@crash_command(
    description="linked list contents",
    arguments=(
        argument(
            "-o",
            dest="offset",
            help="""
            offset from a list entry to the pointer to the next entry, as
            either a number of bytes or a struct_name.member. This can also be
            passed as a positional argument (i.e., without "-o") unless the
            offset could be confused for a kernel virtual address. Defaults to
            0
            """,
        ),
        argument(
            "-s",
            "-S",
            dest="struct",
            metavar="struct_name[.member[,member]]",
            help="""
            type name of list entries. May include one or more comma-separated
            members (each of which can be nested and include array subscripts)
            to limit the output to those members
            """,
        ),
        argument(
            "-l",
            dest="node_offset",
            metavar="OFFSET",
            help="""
            only used with -s. Cannot be used with -o. Offset from a list entry
            to the list_head or similar structure linking entries together, as
            either a number of bytes or a struct_name.member
            """,
        ),
        mutually_exclusive_group(
            argument(
                "-H",
                dest="list_head",
                action="store_true",
                help="""
                treat **start** as the address of an anchor list_head (e.g., a
                standalone LIST_HEAD())
                """,
            ),
            argument(
                "-h",
                dest="embedded_list_head",
                action="store_true",
                help="""
                treat **start** as the address of a list_head embedded in a
                list entry
                """,
            ),
        ),
        argument(
            "-O",
            dest="head_node_offset",
            metavar="OFFSET",
            help="""
            only used with -h. Treat **start** as the address of a structure
            that embeds an anchor list_head at this offset, given as either a
            number of bytes or a struct_name.member
            """,
        ),
        argument(
            "-r",
            dest="reverse",
            action="store_true",
            help="""
            with -h or -H, iterate in reverse order using list_head.prev
            """,
        ),
        argument(
            "-e",
            dest="end",
            type="hexadecimal",
            help="""
            hexadecimal address where to stop iterating in addition to the
            defaults of NULL, the start address, the address of the first
            entry, and the address of the current entry
            """,
        ),
        argument(
            "-x",
            dest="integer_base",
            action="store_const",
            const=16,
            help="output integers in hexadecimal format regardless of the default",
        ),
        argument(
            "-d",
            dest="integer_base",
            action="store_const",
            const=10,
            help="output integers in decimal format regardless of the default",
        ),
        # offset can be passed positionally either before or after start.
        # argparse can't express this, so we add this hidden positional
        # argument and figure out which is actually which manually.
        argument(
            "posarg",
            nargs="?",
            help=argparse.SUPPRESS,
        ),
        argument(
            "start",
            help="hexadecimal address of first list entry (but see -H, -h, and -O)",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_list(
    prog: Program,
    name: str,
    args: argparse.Namespace,
    *,
    parser: argparse.ArgumentParser,
    **kwargs: Any,
) -> None:
    # In crash, if -H is not given, -r implies -h. I don't know whether this is
    # intentional or whether anyone relies on it, so we preserve that.
    if args.reverse and not args.list_head:
        args.embedded_list_head = True

    posargs = []
    if args.posarg is not None:
        posargs.append(args.posarg)
    posargs.append(args.start)
    args.start = None

    number_posargs = []
    for i, posarg in enumerate(posargs):
        # If the argument contains ".", then it must be the offset.
        if "." in posarg:
            if args.offset is not None:
                parser.error("offset given multiple times")
            args.offset = posarg
            continue

        try:
            # If the argument is numeric, it could be either the offset or the
            # start. Save it so we can decide later.
            #
            # Note that a number that is valid in base 10 is also valid in base
            # 16, so this covers both.
            int(posarg, 16)
            number_posargs.append(posarg)
        except ValueError:
            # The argument isn't numeric and doesn't contain ".". It must be a
            # symbol.
            if args.start is None:
                args.start = posarg
            else:
                parser.error("start given multiple times")

    if len(number_posargs) == 2:
        # There were two numeric positional arguments. If a numeric argument is
        # a valid virtual address, assume that it is the start, and otherwise
        # assume that it is the offset.
        is_virt_addr = []
        for posarg in number_posargs:
            try:
                prog.read_word(int(posarg, 16))
            except FaultError:
                is_virt_addr.append(False)
            else:
                is_virt_addr.append(True)

        if is_virt_addr[0] and is_virt_addr[1]:
            # Two valid virtual addresses were given.
            if args.offset is not None:
                # -o was also given, so assume both were intended as the start.
                parser.error("start given multiple times")
            parser.error(
                f"ambiguous arguments: {number_posargs[0]!r} and {number_posargs[1]!r}: -o is required"
            )

        if (not is_virt_addr[0] and not is_virt_addr[1]) or args.offset is not None:
            # Neither is a valid virtual address, or one is a valid virtual
            # address and -o was also given.
            parser.error("offset given multiple times")

        assert args.start is None
        if is_virt_addr[0]:
            args.start, args.offset = number_posargs
        else:
            args.offset, args.start = number_posargs
    elif len(number_posargs) == 1:
        # There was one numeric argument. If we don't know the start yet, it
        # must be the start. Otherwise, it must be the offset.
        if args.start is None:
            args.start = number_posargs[0]
        else:
            if args.offset is not None:
                parser.error("offset given multiple times")
            args.offset = number_posargs[0]

    if args.start is None:
        parser.error("start address is required")

    if args.offset is None:
        next_offset_arg = None
    else:
        next_offset_arg = _parse_type_offset_arg(args.offset)

    if args.struct is None:
        entry_name = None
        entry_members: Sequence[str] = ()
    else:
        entry_name, entry_members = _parse_type_name_and_members(args.struct)

    if args.head_node_offset is None:
        head_node_offset_arg = None
    else:
        if not args.embedded_list_head:
            parser.error("-O can only be used with -h")
        head_node_offset_arg = _parse_type_offset_arg(args.head_node_offset)

    if args.node_offset is None:
        node_offset_arg = None
    else:
        if args.struct is None:
            parser.error("-l can only be used with -s/-S")
        if args.offset is not None:
            parser.error("-l not allowed with -o/offset")
        node_offset_arg = _parse_type_offset_arg(args.node_offset)

    if args.drgn:
        return _list_drgn_option(
            prog,
            args,
            next_offset_arg=next_offset_arg,
            head_node_offset_arg=head_node_offset_arg,
            node_offset_arg=node_offset_arg,
            entry_name=entry_name,
            entry_members=entry_members,
        )

    start = _resolve_addr_or_sym(prog, _addr_or_sym(args.start))

    if entry_name is None:
        entry_type = None
    else:
        entry_type = _guess_type(prog, entry_name)

    next_offset = _resolve_type_offset_arg(prog, next_offset_arg, entry_type)
    head_node_offset = _resolve_type_offset_arg(
        prog,
        head_node_offset_arg,
        # entry_type isn't expected to match for -O, but pass it anyways in
        # case it does.
        entry_type,
    )
    node_offset = _resolve_type_offset_arg(prog, node_offset_arg, entry_type)

    list_head_offset = 0
    if args.list_head or args.embedded_list_head:
        list_head_offset = next_offset
        # Note: -O -r on crash prints only the first list entry and then stops.
        # That's probably a bug, so we implement the more obvious behavior.
        next_offset = prog.address_size() if args.reverse else 0
        if args.list_head or args.head_node_offset is not None:
            if not args.end:
                args.end = start + head_node_offset
            start = prog.read_word(start + head_node_offset + next_offset)
            if start == args.end:
                print("(empty)")
                return
        else:
            start += list_head_offset

    entry_offset = list_head_offset + node_offset

    ptr = start
    end = {start}
    if args.end is not None:
        end.add(args.end)
    seen: Set[int] = set()
    format_options = _object_format_options(prog, args.integer_base)
    while True:
        # Note: with -H or -h, -l prints the address of the list node, but -o
        # prints the address of the entry.
        print(f"{ptr - list_head_offset:x}")
        if entry_type is not None:
            entry = Object(prog, entry_type, address=ptr - entry_offset)
            if entry_members:
                for member in entry_members:
                    print(
                        f"  {member} = {entry.subobject_(member).format_(**format_options)}"
                    )
            else:
                print(entry.format_(**format_options))

        if ptr - list_head_offset in seen:
            print(f"\nlist: duplicate list entry: {ptr:x}")
            break

        next_ptr = prog.read_word(ptr + next_offset)
        if not next_ptr:
            if args.list_head or args.embedded_list_head:
                print("\ninvalid list entry: 0")
            break
        elif next_ptr in end or next_ptr == ptr:
            break

        if not seen:
            end.add(next_ptr)
        seen.add(ptr - list_head_offset)

        ptr = next_ptr


_TreeType = Literal[
    "rbtree",
    "radix",
    "xarray",
    "maple",
]


def _find_tree_type(name: str) -> _TreeType:
    tree_types = [type for type in typing.get_args(_TreeType) if type.startswith(name)]
    if len(tree_types) != 1:
        raise CommandArgumentError(f"invalid tree type: {name}")
    return tree_types[0]


def _tree_drgn_option(
    prog: Program,
    args: argparse.Namespace,
    *,
    tree_type: _TreeType,
    root_offset_arg: Union[None, int, Tuple[str, str]],
    node_offset_arg: Union[None, int, Tuple[str, str]],
    entry_name: Optional[str],
    entry_members: Sequence[str],
) -> None:
    code = CrashDrgnCodeBuilder(prog)

    if tree_type == "rbtree":
        root_type_name = "struct rb_root"
        helper_module = "drgn.helpers.linux.rbtree"
    elif tree_type == "radix":
        try:
            root_type_name = "struct radix_tree_root"
            prog.type(root_type_name)
        except LookupError:
            root_type_name = "struct xarray"
        helper_module = "drgn.helpers.linux.radixtree"
        helper_name = "radix_tree_for_each"
        target = "index, entry"
    elif tree_type == "xarray":
        root_type_name = "struct xarray"
        helper_module = "drgn.helpers.linux.xarray"
        helper_name = "xa_for_each"
        target = "index, entry"
    elif tree_type == "maple":
        root_type_name = "struct maple_tree"
        helper_module = "drgn.helpers.linux.mapletree"
        helper_name = "mt_for_each"
        target = "first_index, last_index, entry"
    else:
        assert_never(tree_type)

    start_type = None
    if isinstance(root_offset_arg, tuple):
        try:
            start_type = _guess_type(prog, root_offset_arg[0])
        except LookupError:
            start_type = None
            start_type_name = "struct " + root_offset_arg[0]
        else:
            start_type_name = start_type.type_name()

    if args.start[0] == "addr":
        address = args.start[1]
        code.add_from_import("drgn", "Object")
        if isinstance(root_offset_arg, int):
            address_arg = f"{hex(address)} + {root_offset_arg}"
        elif isinstance(root_offset_arg, tuple):
            code.append(f"address = {hex(address)}\n")
            code.add_from_import("drgn", "offsetof")
            code.append(
                f'address += offsetof(prog.type("{start_type_name}"), "{root_offset_arg[1]}")\n'
            )
            address_arg = "address"
        else:
            address_arg = hex(address)
        code.append(f'root = Object(prog, "{root_type_name} *", {address_arg})\n')
    else:
        symbol_name = args.start[1]
        if root_offset_arg is None and _prefer_object_lookup(
            prog, root_type_name, symbol_name
        ):
            code.append(f"root = prog[{_repr_black(symbol_name)}].address_of_()\n")
        elif (
            isinstance(root_offset_arg, tuple)
            and start_type is not None
            and _prefer_object_lookup(prog, start_type_name, symbol_name)
            and typeof_member(start_type, root_offset_arg[1]).type_name()
            == root_type_name
        ):
            code.append(
                f"root = prog[{_repr_black(symbol_name)}].{root_offset_arg[1]}.address_of_()\n"
            )
        else:
            code.append(f"address = prog.symbol({_repr_black(symbol_name)}).address")

            if isinstance(root_offset_arg, int):
                code.append(f" + {root_offset_arg}\n")
            elif isinstance(root_offset_arg, tuple):
                code.add_from_import("drgn", "offsetof")
                code.append(
                    f'\naddress += offsetof(prog.type("{start_type_name}"), "{root_offset_arg[1]}")\n'
                )
            else:
                code.append("\n")

            code.add_from_import("drgn", "Object")
            code.append(f'root = Object(prog, "{root_type_name} *", address)\n')

    if entry_name is None:
        entry_type_name = None
    else:
        entry_type_name = _guess_type_name(prog, entry_name)

    have_loop_body = False

    if tree_type == "rbtree":
        if isinstance(node_offset_arg, tuple):
            if node_offset_arg[0] == entry_name:
                for_each_entry_type_name = entry_type_name
            else:
                for_each_entry_type_name = _guess_type_name(prog, node_offset_arg[0])
            helper_name = (
                "rbtree_inorder_for_each_entry"
                if args.linear
                else "rbtree_preorder_for_each_entry"
            )
            code.append(
                f"""\
for entry in {helper_name}(
    "{for_each_entry_type_name}", root, "{node_offset_arg[1]}"
):
"""
            )
            if (
                entry_type_name is not None
                and entry_type_name != for_each_entry_type_name
            ):
                code.add_from_import("drgn", "cast")
                code.append(f'    entry = cast("{entry_type_name} *", entry)\n')
                have_loop_body = True
        else:
            helper_name = (
                "rbtree_inorder_for_each" if args.linear else "rbtree_preorder_for_each"
            )
            code.append(f"for node in {helper_name}(root):\n")
            if isinstance(node_offset_arg, int):
                code.add_from_import("drgn", "cast")
                code.append(f'    entry = cast("void *", node) - {node_offset_arg}\n')
                to_cast = "entry"
                have_loop_body = True
            else:
                to_cast = "node"
            if entry_type_name is not None:
                code.add_from_import("drgn", "cast")
                code.append(f'    entry = cast("{entry_type_name} *", {to_cast})\n')
                have_loop_body = True
    else:
        code.append(f"for {target} in {helper_name}(root):\n")
        if entry_type_name is not None:
            code.add_from_import("drgn", "cast")
            code.append(f'    entry = cast("{entry_type_name} *", entry)\n')
            have_loop_body = True

    code.add_from_import(helper_module, helper_name)

    if entry_members:
        _append_get_members(code, "entry", entry_members, indent="    ")
        have_loop_body = True

    if not have_loop_body:
        code.append("    pass\n")

    code.print()


@crash_command(
    description="red-black tree, radix tree, XArray, or maple tree contents",
    arguments=(
        argument(
            "-t",
            dest="type",
            metavar="{rbtree,radix,xarray,maple}",
            default="rbtree",
            help="""
            type of tree. This may also be abbreviated: "rb" for "rbtree", "ra"
            for "radix", "x" for "xarray", or "m" for "maple". Defaults to
            rbtree
            """,
        ),
        argument(
            "-r",
            dest="root_offset",
            metavar="OFFSET",
            help="""
            treat **start** as the address of a structure containing the tree
            root at the given offset as either a number of bytes or a
            struct_name.member
            """,
        ),
        argument(
            "-o",
            dest="node_offset",
            metavar="OFFSET",
            help="""
            for red-black trees, the offset of the rb_node in its containing
            structure as either a number of bytes or a struct_name.member.
            Defaults to 0
            """,
        ),
        argument(
            "-s",
            "-S",
            dest="struct",
            metavar="struct_name[.member[,member]]",
            help="""
            type name of tree entries. May include one or more comma-separated
            members (each of which can be nested and include array subscripts)
            to limit the output to those members
            """,
        ),
        argument(
            "-l",
            dest="linear",
            action="store_true",
            help="""
            for red-black trees, print entries in sorted order instead of the
            default pre-order traversal
            """,
        ),
        argument(
            "-x",
            dest="integer_base",
            action="store_const",
            const=16,
            help="output integers in hexadecimal format regardless of the default",
        ),
        argument(
            "-d",
            dest="integer_base",
            action="store_const",
            const=10,
            help="output integers in decimal format regardless of the default",
        ),
        argument(
            "start",
            type="addr_or_sym",
            help="""
            address or symbol name of the rb_root, radix_tree_root, xarray, or
            maple_tree (or the structure containing it if **-r** is used)
            """,
        ),
        drgn_argument,
    ),
)
def _crash_cmd_tree(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    tree_type = _find_tree_type(args.type)

    if args.root_offset is None:
        root_offset_arg = None
    else:
        root_offset_arg = _parse_type_offset_arg(args.root_offset)

    if args.node_offset is None:
        node_offset_arg = None
    else:
        if tree_type != "rbtree":
            raise CommandArgumentError(f"-o is not applicable to {tree_type}")
        node_offset_arg = _parse_type_offset_arg(args.node_offset)

    if args.struct is None:
        entry_name = None
        entry_members: Sequence[str] = ()
    else:
        entry_name, entry_members = _parse_type_name_and_members(args.struct)

    if args.linear and tree_type != "rbtree":
        raise CommandArgumentError(f"-l is not applicable to {tree_type}")

    if args.drgn:
        _tree_drgn_option(
            prog,
            args,
            tree_type=tree_type,
            root_offset_arg=root_offset_arg,
            node_offset_arg=node_offset_arg,
            entry_name=entry_name,
            entry_members=entry_members,
        )
        return

    if entry_name is None:
        entry_type = None
    else:
        entry_type = _guess_type(prog, entry_name)

    start = _resolve_addr_or_sym(prog, args.start) + _resolve_type_offset_arg(
        prog, root_offset_arg
    )

    format_options = _object_format_options(prog, args.integer_base)

    def print_entry(address: int) -> None:
        print(f"{address:x}")
        if entry_type is not None:
            entry = Object(prog, entry_type, address=address)
            if entry_members:
                for member in entry_members:
                    print(
                        f"  {member} = {entry.subobject_(member).format_(**format_options)}"
                    )
            else:
                print(entry.format_(**format_options))

    if tree_type == "rbtree":
        node_offset = _resolve_type_offset_arg(prog, node_offset_arg, entry_type)
        root = Object(prog, "struct rb_root *", start)
        helper = rbtree_inorder_for_each if args.linear else rbtree_preorder_for_each
        for node in helper(root):
            print_entry(node.value_() - node_offset)
    elif tree_type == "radix":
        try:
            radix_tree_root_type = prog.type("struct xarray *")
        except LookupError:
            radix_tree_root_type = prog.type("struct radix_tree_root *")
        root = Object(prog, radix_tree_root_type, start)
        for _, entry in radix_tree_for_each(root):
            print_entry(entry.value_())
    elif tree_type == "xarray":
        root = Object(prog, "struct xarray *", start)
        for _, entry in xa_for_each(root):
            print_entry(entry.value_())
    elif tree_type == "maple":
        root = Object(prog, "struct maple_tree *", start)
        for _, _, entry in mt_for_each(root):
            print_entry(entry.value_())
    else:
        assert_never(tree_type)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Memory
------

The ``drgn.helpers.common.memory`` module provides helpers for working with memory and addresses.
"""

import dataclasses
import operator
from typing import Any, Dict, Iterator, Literal, Optional, Protocol

import drgn
from drgn import IntegerLike, PlatformFlags, Program, Symbol, SymbolKind
from drgn.helpers.common.prog import takes_program_or_default

__all__ = (
    "identify_address",
    "identify_address_all",
    "print_annotated_memory",
)


_SYMBOL_KIND_STR = {
    SymbolKind.OBJECT: "object symbol",
    SymbolKind.FUNC: "function symbol",
}


@takes_program_or_default
def identify_address(
    prog: Program, addr: IntegerLike, *, cache: Optional[Dict[Any, Any]] = None
) -> Optional[str]:
    """
    Try to identify what an address refers to.

    For all programs, this will identify addresses as follows:

    * Object symbols (e.g., addresses in global variables):
      ``object symbol: {symbol_name}+{hex_offset}`` (where ``hex_offset`` is
      the offset from the beginning of the symbol in hexadecimal).
    * Function symbols (i.e., addresses in functions):
      ``function symbol: {symbol_name}+{hex_offset}``.
    * Other symbols: ``symbol: {symbol_name}+{hex_offset}``.

    Additionally, for the Linux kernel, this will identify:

    * Task structures: ``task: {pid} ({comm}) +{hex_offset}`` (where ``pid``
      and ``comm`` identify the task and ``hex_offset`` is the optional offset
      from the beginning of the structure).
    * Task stacks: ``task stack: {pid} ({comm}) +{hex_offset}`` (where
      ``pid`` and ``comm`` identify the task and ``hex_offset`` is the offset
      from the beginning of the stack in hexadecimal).
    * Allocated slab objects: ``slab object: {slab_cache_name}+{hex_offset}``
      (where ``hex_offset`` is the offset from the beginning of the object in
      hexadecimal).
    * Free slab objects: ``free slab object: {slab_cache_name}+{hex_offset}``.
    * Slab objects with SLOB: ``unknown slab object``.
    * Page structures: ``page: pfn {pfn} +{hex_offset}`` (where ``pfn`` is the
      page frame number and ``hex_offset`` is the optional offset from the
      beginning of the structure).
    * Vmap addresses (e.g., vmalloc, ioremap):
      ``vmap: {hex_start_address}-{hex_end_address}``. If the function that
      allocated the vmap is known, this also includes
      ``caller {function_name}+{hex_offset}``.

    This may recognize other types of addresses in the future.

    To get more information instead of just a string, use
    :func:`identify_address_all()`.

    :param addr: ``void *``
    :param cache: Opaque cache used to amortize expensive lookups. If you're
        going to call this function many times in a short period, create an
        empty dictionary and pass the same dictionary as *cache* to each call.
        Don't reuse it indefinitely or you may get stale results.
    :return: Identity as string, or ``None`` if the address is unrecognized.
    """
    for identified in identify_address_all(prog, addr, cache=cache):
        return str(identified)
    return None


@takes_program_or_default
def identify_address_all(
    prog: Program, addr: IntegerLike, *, cache: Optional[Dict[Any, Any]] = None
) -> Iterator["IdentifiedAddress"]:
    """
    Identify everything an address refers to.

    This is a more programmatic variant of :func:`identify_address()` that
    provides the following additional information:

    * Instead of strings, it yields :class:`IdentifiedAddress` instances which
      have attributes describing the identity of the address.
    * If the address can be identified in multiple ways, it yields each one.
      For example, a pointer to a Linux kernel ``task_struct`` can be
      identified as both a task and a slab object.

    For all programs, this can yield:

    * :class:`IdentifiedSymbol`

    Additionally, for the Linux kernel, this can yield:

    * :class:`~drgn.helpers.linux.common.IdentifiedTaskStruct`
    * :class:`~drgn.helpers.linux.common.IdentifiedTaskStack`
    * :class:`~drgn.helpers.linux.common.IdentifiedSlabObject`
    * :class:`~drgn.helpers.linux.common.IdentifiedVmap`
    * :class:`~drgn.helpers.linux.common.IdentifiedPage`

    :param addr: ``void *``
    :param cache: Opaque cache used to amortize expensive lookups. If you're
        going to call this function many times in a short period, create an
        empty dictionary and pass the same dictionary as *cache* to each call.
        Don't reuse it indefinitely or you may get stale results.
    :return: Iterator of identities, from most specific to least specific. If
        the address is unrecognized, this is empty.
    """
    is_linux_kernel = bool(prog.flags & drgn.ProgramFlags.IS_LINUX_KERNEL)
    if is_linux_kernel:
        from drgn.helpers.linux.common import (
            _identify_kernel_address,
            _identify_kernel_symbol,
        )

    addr = operator.index(addr)
    # Check if address is of a symbol:
    symbol: Optional[Symbol]
    try:
        symbol = prog.symbol(addr)
    except LookupError:  # not a symbol
        symbol = None
    else:
        if is_linux_kernel:
            yield from _identify_kernel_symbol(prog, addr, symbol)

        yield IdentifiedSymbol(addr, symbol)

    if is_linux_kernel:
        yield from _identify_kernel_address(prog, addr, bool(symbol), cache)


class IdentifiedAddress(Protocol):
    """Address that was identified by :func:`identify_address_all()`."""

    address: int
    """Address passed to :func:`identify_address_all()`."""

    def __str__(self) -> str:
        """
        Get a human-readable description of the identity.

        This is the same string that is returned from
        :func:`identify_address()`.

        >>> identity
        IdentifiedSymbol(address=18446744071889293504, symbol=Symbol(name='init_task', address=0xffffffff938110c0, size=0x36c0, binding=<SymbolBinding.GLOBAL: 2>, kind=<SymbolKind.OBJECT: 1>))
        >>> str(identity)
        'object symbol: init_task+0x0'
        """
        ...


@dataclasses.dataclass
class IdentifiedSymbol:
    """:class:`IdentifiedAddress` for an address in the symbol table."""

    address: int

    symbol: Symbol
    """Symbol containing the address."""

    def __str__(self) -> str:
        symbol_kind = _SYMBOL_KIND_STR.get(self.symbol.kind, "symbol")
        return f"{symbol_kind}: {self.symbol.name}+{hex(self.address - self.symbol.address)}"


@takes_program_or_default
def print_annotated_memory(
    prog: Program, address: IntegerLike, size: IntegerLike, physical: bool = False
) -> None:
    """
    Print the contents of a range of memory, annotating values that can be
    identified.

    Currently, this will identify any addresses in the memory range with
    :func:`~drgn.helpers.common.memory.identify_address()`.

    See :func:`~drgn.helpers.common.stack.print_annotated_stack()` for a
    similar function that annotates stack traces.

    >>> print_annotated_memory(0xffffffff963eb200, 56)
    ADDRESS           VALUE
    ffffffff963eb200: 00000000000000b8
    ffffffff963eb208: 000000000000a828
    ffffffff963eb210: 0000000000000000
    ffffffff963eb218: ffff8881042948e0 [slab object: mnt_cache+0x20]
    ffffffff963eb220: ffff88810074a540 [slab object: dentry+0x0]
    ffffffff963eb228: ffff8881042948e0 [slab object: mnt_cache+0x20]
    ffffffff963eb230: ffff88810074a540 [slab object: dentry+0x0]

    :param address: Starting address.
    :param size: Number of bytes to read.
    :param physical: Whether *address* is a physical memory address. If
        ``False``, then it is a virtual memory address.
    """
    address = operator.index(address)
    mem = prog.read(address, size, physical)

    # The platform must be known if we were able to read memory.
    assert prog.platform is not None

    byteorder: Literal["little", "big"]
    if prog.platform.flags & PlatformFlags.IS_LITTLE_ENDIAN:
        byteorder = "little"
    else:
        byteorder = "big"

    word_size = prog.address_size()
    if word_size == 8:
        line_format = "{:016x}: {:016x}{}"
        print("ADDRESS           VALUE")
    else:
        line_format = "{:08x}: {:08x}{}"
        print("ADDRESS   VALUE")

    cache: Dict[Any, Any] = {}
    for offset in range(0, len(mem), word_size):
        value = int.from_bytes(mem[offset : offset + word_size], byteorder)
        identified = identify_address(prog, value, cache=cache)
        if identified is None:
            identified = ""
        else:
            identified = f" [{identified}]"
        print(line_format.format(address + offset, value, identified))



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import logging
import sys

import drgn
import drgn.cli
from drgn.commands._crash import _cmd_crash


def _main() -> None:
    drgn.cli._set_log_handler()

    parser = argparse.ArgumentParser(description="Run drgn in crash compatibility mode")
    parser.add_argument(
        "vmlinux",
        nargs="?",
        help="path to kernel image (vmlinux) (default: find automatically)",
    )
    parser.add_argument(
        "core",
        nargs="?",
        help="path to kernel core dump (default: debug the running kernel)",
    )
    args = parser.parse_args()

    posargs = []
    if args.vmlinux is not None:
        posargs.append(args.vmlinux)
    if args.core is not None:
        posargs.append(args.core)

    prog = drgn.Program()

    found_core = False
    debug_info_paths = []
    for posarg in posargs:
        if drgn.cli._identify_script(posarg) == "core":
            if found_core:
                sys.exit("too many core dump arguments")
            else:
                prog.set_core_dump(posarg)
                found_core = True
        else:
            debug_info_paths.append(posarg)
    if not found_core:
        drgn.cli._set_kernel_with_sudo_fallback(prog)

    try:
        prog.load_debug_info(debug_info_paths, default=True)
    except drgn.MissingDebugInfoError as e:
        try:
            main_module = prog.main_module()
            critical = main_module.wants_debug_file() or main_module.wants_loaded_file()
        except LookupError:
            critical = True
        drgn.cli.logger.log(logging.CRITICAL if critical else logging.WARNING, "%s", e)

    sys.displayhook = drgn.cli._displayhook
    drgn.set_default_prog(prog)

    _cmd_crash(
        prog,
        "crash",
        "",
        parser=None,  # type: ignore[arg-type]  # The parser isn't needed.
        globals=drgn.cli.default_globals(prog),
    )


if __name__ == "__main__":
    _main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Stack
-----

The ``drgn.helpers.common.stack`` module provides helpers for working with
stack traces.
"""

from typing import Any, Dict

from drgn import Architecture, FaultError, PlatformFlags, Program, StackTrace
from drgn.helpers.common.memory import identify_address
from drgn.helpers.common.prog import takes_program_or_default

__all__ = (
    "print_annotated_stack",
    "print_registers",
)


def print_annotated_stack(trace: StackTrace) -> None:
    """
    Print the contents of stack memory in a stack trace, annotating values that
    can be identified.

    Currently, this will identify any addresses on the stack with
    :func:`~drgn.helpers.common.memory.identify_address()`.

    See :func:`~drgn.helpers.common.memory.print_annotated_memory()` for a
    similar function that annotates arbitrary memory ranges.

    >>> print_annotated_stack(stack_trace(1))
    STACK POINTER     VALUE
    [stack frame #0 at 0xffffffff8dc93c41 (__schedule+0x429/0x488) in context_switch at ./kernel/sched/core.c:5209:2 (inlined)]
    [stack frame #1 at 0xffffffff8dc93c41 (__schedule+0x429/0x488) in __schedule at ./kernel/sched/core.c:6521:8]
    ffffa903c0013d28: ffffffff8d8497bf [function symbol: __flush_tlb_one_user+0x5]
    ffffa903c0013d30: 000000008d849eb5
    ffffa903c0013d38: 0000000000000001
    ffffa903c0013d40: 0000000000000004
    ffffa903c0013d48: efdea37bb7cb1f00
    ffffa903c0013d50: ffff926641178000 [slab object: task_struct+0x0]
    ffffa903c0013d58: ffff926641178000 [slab object: task_struct+0x0]
    ffffa903c0013d60: ffffa903c0013e10
    ffffa903c0013d68: ffff926641177ff0 [slab object: mm_struct+0x70]
    ffffa903c0013d70: ffff926641178000 [slab object: task_struct+0x0]
    ffffa903c0013d78: ffff926641178000 [slab object: task_struct+0x0]
    ffffa903c0013d80: ffffffff8dc93d29 [function symbol: schedule+0x89]
    ...

    :param trace: Stack trace to print.
    """
    prog = trace.prog

    # platform must be known if there is a stack trace
    assert prog.platform is not None

    if prog.platform.flags & PlatformFlags.IS_LITTLE_ENDIAN:
        byteorder = "little"
    else:
        byteorder = "big"

    word_size = prog.address_size()
    if word_size == 8:
        line_format = "{:016x}: {:016x}{}"
        print("STACK POINTER     VALUE")
    else:
        line_format = "{:08x}: {:08x}{}"
        print("STACK     VALUE\nPOINTER")

    cache: Dict[Any, Any] = {}
    start = 0
    while start < len(trace):
        # Find the bounds of this stack. Our heuristics for the end of the
        # stack are:
        #
        # 1. An interrupted frame.
        # 2. A frame with a stack pointer less than the previous frame's stack
        #    pointer (since the stack grows down in all of the architectures we
        #    support).
        # 3. A frame with a stack pointer much greater than the previous
        #    frame's stack pointer. Our arbitrary threshold is 128 MB. (Linux
        #    kernel stacks are at most 16 KB as of Linux 6.8, and userspace
        #    stacks are limited to 8 MB by default, so this threshold could be
        #    adjusted if needed.)
        end = start + 1
        while (
            end < len(trace)
            and not trace[end].interrupted
            and 0 <= trace[end].sp - trace[end - 1].sp <= 128 * 1024 * 1024
        ):
            end += 1

        # Gather the frames for this stack.
        frames = [trace[i] for i in range(start, end)]
        frames_addrs = [frame.sp for frame in frames]

        start_addr = frames_addrs[0]
        end_addr = frames_addrs[-1] + word_size - 1
        stack_size = end_addr - start_addr + 1

        try:
            stack_bytes = prog.read(start_addr, stack_size)
        except FaultError:
            # Couldn't read the stack. Just print the frames.
            for frame in frames:
                print(f"[stack frame {frame}]")

            start = end
            continue

        frame_ind = 0

        for offset in range(0, len(stack_bytes), word_size):
            addr = start_addr + offset

            word_bytes = stack_bytes[offset : offset + word_size]
            word_val = int.from_bytes(
                word_bytes,
                # The byteorder parameter is annotated as
                # Literal['little', 'big'], but mypy infers that our byteorder
                # variable is str.
                byteorder=byteorder,  # type: ignore[arg-type]
            )

            # There may be multiple frames matching this address (usually
            # because of inline frames).
            while frame_ind < len(frames_addrs) and addr == frames_addrs[frame_ind]:
                frame = frames[frame_ind]
                print(f"[stack frame {frame}]")
                frame_ind += 1

            identified = identify_address(prog, word_val, cache=cache)
            if identified is None:
                identified = ""
            else:
                identified = f" [{identified}]"
            print(line_format.format(addr, word_val, identified))

        start = end


@takes_program_or_default
def print_registers(prog: Program, regs: Dict[str, int], indent: str = "    ") -> None:
    """
    Print a CPU register dump, in a format similar to that of
    :manpage:`crash(8)`

    :param regs: a dictionary of registers, named in a similar way to the
      dictionary returned by :py:meth:`drgn.StackTrace.registers()`.
    :param indent: a string prepended to each line of output
    """
    if not prog.platform:
        raise RuntimeError("Unknown platform & architecture")

    widths = {}
    formats = {}
    if prog.platform.arch == Architecture.X86_64:
        rows = [
            ("rip", "rsp", "rflags"),
            ("rax", "rbx", "rcx"),
            ("rdx", "rsi", "rdi"),
            ("rbp", "r8", "r9"),
            ("r10", "r11", "r12"),
            ("r13", "r14", "r15"),
            ("cs", "ss"),
        ]
        widths = {"cs": 4, "ss": 4, "rflags": 8}
    elif prog.platform.arch == Architecture.AARCH64:
        rows = [
            ("pc", "lr", "sp"),
            ("x29", "x28", "x27"),
            ("x26", "x25", "x24"),
            ("x23", "x22", "x21"),
            ("x20", "x19", "x18"),
            ("x17", "x16", "x15"),
            ("x14", "x13", "x12"),
            ("x11", "x10", "x9"),
            ("x8", "x7", "x6"),
            ("x5", "x4", "x3"),
            ("x2", "x1", "x0"),
        ]
    elif prog.platform.arch == Architecture.ARM:
        rows = [
            ("pc", "lr", "sp", "fp"),
            ("r10", "r9", "r8"),
            ("r7", "r6", "r5", "r4"),
            ("r3", "r2", "r1", "r0"),
        ]
    elif prog.platform.arch == Architecture.S390X:
        rows = [
            ("pswm", "pswa", "r0"),
            ("r1", "r3", "r3"),
            ("r4", "r5", "r6"),
            ("r7", "r8", "r9"),
            ("r10", "r11", "r12"),
            ("r13", "r14", "r15"),
        ]
    elif prog.platform.arch == Architecture.PPC64:
        rows = [
            ("r0", "r1", "r2"),
            ("r3", "r4", "r5"),
            ("r6", "r7", "r8"),
            ("r9", "r10", "r11"),
            ("r12", "r13", "r14"),
            ("r15", "r16", "r17"),
            ("r18", "r19", "r20"),
            ("r21", "r22", "r23"),
            ("r24", "r25", "r26"),
            ("r27", "r28", "r29"),
            ("r30", "r31"),
            # This departs from the crash format significantly. Crash recovers
            # registers like CTR, XER, LR, and a few others. However, drgn
            # doesn't provide these in stack frame registers. They are available
            # in struct pt_regs.
            ("cr0", "cr1", "cr2", "cr3"),
            ("cr4", "cr5", "cr6", "cr7"),
        ]
        widths = {f"cr{i}": 4 for i in range(8)}
        formats = {f"cr{i}": "b" for i in range(8)}
    else:
        raise RuntimeError(f"Unsupported architecture: {prog.platform.arch}")

    default_width = 16 if prog.platform.flags & PlatformFlags.IS_64_BIT else 8

    for row in rows:
        row_text = []
        for i, reg in enumerate(row):
            width = widths.get(reg, default_width)
            if reg in regs:
                value = regs[reg]
                fmt = formats.get(reg, "x")
                row_text.append(f"{reg.upper():>3s}: {value:0{width}{fmt}}")
            else:
                row_text.append(f"{reg.upper():>3s}: {'?' * width}")

        print(f"{indent}{'  '.join(row_text)}")



# Copyright (c) 2025 Oracle and/or its affiliates
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Implements the crash "whatis" command for drgn."""

import argparse
from typing import Any

from drgn import ObjectNotFoundError, Program, sizeof
from drgn.commands import CommandError, argument, drgn_argument
from drgn.commands._crash.common import CrashDrgnCodeBuilder, crash_command


@crash_command(
    name="whatis",
    description="search symbol table for data or type information",
    long_description="""
    Displays the definition of structures, unions, typedefs or text/data symbols.
    """,
    arguments=(
        argument(
            "name",
            help="struct/union/typedef/enum/type name or symbol name",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_whatis(
    prog: Program, cmd: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    target = args.name

    typ = None
    try:
        # Try as-is (typedefs and names with prefixes).
        typ = prog.type(target)
    except LookupError:
        # Try C prefixes 'struct', 'union' and 'enum' if not provided.
        if not target.startswith(("struct ", "union ", "enum ")):
            for prefix in ("struct ", "union ", "enum "):
                try:
                    typ = prog.type(prefix + target)
                    break
                except LookupError:
                    continue

    if typ is not None:
        if args.drgn:
            code = CrashDrgnCodeBuilder(prog)
            code.append(f'typ = prog.type("{typ.type_name()}")\n')
            code.append(
                """try:
    size = sizeof(typ)
except TypeError:
    # Type doesn't have size
    pass\n"""
            )
            code.print()
        else:
            try:
                print(f"{typ};\nSIZE: {sizeof(typ)}")
            except TypeError:
                print(f"{typ};")
        return

    # Otherwise, try to resolve as variable or function.
    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        code.append(f'typ = prog["{target}"].type_\n')
        code.print()
        return

    try:
        obj = prog[target]
    except ObjectNotFoundError:
        raise CommandError(f"unknown type or symbol: {target}")

    decl = obj.type_.variable_declaration(target)
    print(f"{decl};")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
from typing import List


class MultilineHelpFormatter(argparse.HelpFormatter):
    """
    argparse help formatter that supports multiple paragraphs/line blocks and
    doesn't collapse consecutive whitespace.

    '''
    This is a paragraph, which will
    be wrapped.

    |This is a line block.
    |It will not be wrapped.
    '''

    See :func:`_drgn_util.multilinewrap.multiline_wrap()` for details.
    """

    def _split_lines(self, text: str, width: int) -> List[str]:
        # argparse makes a point of not importing textwrap unless necessary, so
        # we do the same.
        from _drgn_util.multilinewrap import multiline_wrap

        return multiline_wrap(text, width, indent="")

    def _fill_text(self, text: str, width: int, indent: str) -> str:
        from _drgn_util.multilinewrap import multiline_fill

        return multiline_fill(text, width, indent=indent)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
XArrays
-------

The ``drgn.helpers.linux.xarray`` module provides helpers for working with the
`XArray <https://docs.kernel.org/core-api/xarray.html>`_ data structure from
:linux:`include/linux/xarray.h`.

.. note::

    XArrays were introduced in Linux 4.20 as a replacement for `radix trees`_.
    To make it easier to work with data structures that were changed from a
    radix tree to an XArray (like ``struct address_space::i_pages``), drgn
    treats XArrays and radix trees interchangeably in some cases.

    Specifically, :func:`~drgn.helpers.linux.xarray.xa_load()` is equivalent to
    :func:`~drgn.helpers.linux.radixtree.radix_tree_lookup()`, and
    :func:`~drgn.helpers.linux.xarray.xa_for_each()` is equivalent to
    :func:`~drgn.helpers.linux.radixtree.radix_tree_for_each()`, except that
    the radix tree helpers assume ``advanced=False``. (Therefore,
    :func:`~drgn.helpers.linux.xarray.xa_load()` and
    :func:`~drgn.helpers.linux.xarray.xa_for_each()` also accept a ``struct
    radix_tree_root *``, and
    :func:`~drgn.helpers.linux.radixtree.radix_tree_lookup()` and
    :func:`~drgn.helpers.linux.radixtree.radix_tree_for_each()` also accept a
    ``struct xarray *``.)
"""

from typing import Iterator, Optional, Tuple

from _drgn import _linux_helper_xa_load
from drgn import NULL, IntegerLike, Object, cast

__all__ = (
    "xa_for_each",
    "xa_is_value",
    "xa_is_zero",
    "xa_load",
    "xa_to_value",
)


_XA_ZERO_ENTRY = 1030  # xa_mk_internal(257)


def _xa_is_node(entry_value: int) -> bool:
    return (entry_value & 3) == 2 and entry_value > 4096


def xa_load(xa: Object, index: IntegerLike, *, advanced: bool = False) -> Object:
    """
    Look up the entry at a given index in an XArray.

    >>> entry = xa_load(inode.i_mapping.i_pages.address_of_(), 2)
    >>> cast("struct page *", entry)
    *(struct page *)0xffffed6980306f40 = {
        ...
    }

    :param xa: ``struct xarray *``
    :param index: Entry index.
    :param advanced: Whether to return nodes only visible to the XArray
        advanced API. If ``False``, zero entries (see :func:`xa_is_zero()`)
        will be returned as ``NULL``.
    :return: ``void *`` found entry, or ``NULL`` if not found.
    """
    entry = _linux_helper_xa_load(xa, index)
    if not advanced and entry.value_() == _XA_ZERO_ENTRY:
        return NULL(xa.prog_, "void *")
    return entry


class _XAIteratorNode:
    def __init__(self, node: Object, index: int) -> None:
        self.slots = node.slots
        self.shift = node.shift.value_()
        self.index = index
        self.next_slot = 0


def xa_for_each(xa: Object, *, advanced: bool = False) -> Iterator[Tuple[int, Object]]:
    """
    Iterate over all of the entries in an XArray.

    >>> for index, entry in xa_for_each(inode.i_mapping.i_pages.address_of_()):
    ...     print(index, entry)
    ...
    0 (void *)0xffffed6980356140
    1 (void *)0xffffed6980306f80
    2 (void *)0xffffed6980306f40
    3 (void *)0xffffed6980355b40

    :param xa: ``struct xarray *``
    :param advanced: Whether to return nodes only visible to the XArray
        advanced API. If ``False``, zero entries (see :func:`xa_is_zero()`)
        will be skipped.
    :return: Iterator of (index, ``void *``) tuples.
    """
    prog = xa.prog_

    def should_yield(entry_value: int) -> bool:
        return entry_value != 0

    # This handles three cases:
    #
    # 1. XArrays.
    # 2. Radix trees since Linux kernel commit f8d5d0cc145c ("xarray: Add
    #    definition of struct xarray") (in v4.20) redefined them in terms of
    #    XArrays. These reuse the XArray structures and are close enough to
    #    case 1 that the same code handles both.
    # 3. Radix trees before that commit. These are similar to cases 1 and 2,
    #    but they have different type and member names, use different flags in
    #    the lower bits (see Linux kernel commit 3159f943aafd ("xarray: Replace
    #    exceptional entries") (in v4.20)), and represent sibling entries
    #    differently (see Linux kernel commit 02c02bf12c5d ("xarray: Change
    #    definition of sibling entries") (in v4.20)).
    try:
        entry = xa.xa_head.read_()
    except AttributeError:
        entry = xa.rnode
        node_type = entry.type_
        entry = cast("void *", entry)

        # Return > 0 if radix_tree_is_internal_node(), < 0 if
        # is_sibling_entry(), and 0 otherwise.
        def is_internal(slots: Optional[Object], entry_value: int) -> int:
            if (entry_value & 3) == 1:
                # slots must be a reference object, so address_ is never None.
                if slots is not None and (
                    slots.address_ <= entry_value < slots[len(slots)].address_  # type: ignore[operator]
                ):
                    return -1
                else:
                    return 1
            return 0

        # entry_to_node()
        def to_node(entry_value: int) -> Object:
            return Object(prog, node_type, entry_value - 1)

    else:
        node_type = prog.type("struct xa_node *")

        # Return > 0 if xa_is_node(), < 0 if xa_is_sibling(), and 0 otherwise.
        def is_internal(slots: Optional[Object], entry_value: int) -> int:
            if _xa_is_node(entry_value):
                return 1
            elif (entry_value & 3) == 2 and entry_value < 256:
                return -1
            else:
                return 0

        # xa_to_node()
        def to_node(entry_value: int) -> Object:
            return Object(prog, node_type, entry_value - 2)

        if not advanced:
            # We're intentionally redefining should_yield() for this case.
            def should_yield(entry_value: int) -> bool:  # noqa: F811
                return entry_value != 0 and entry_value != _XA_ZERO_ENTRY

    entry_value = entry.value_()
    internal = is_internal(None, entry_value)
    if internal > 0:
        stack = [_XAIteratorNode(to_node(entry_value), 0)]
    else:
        if internal == 0 and should_yield(entry_value):
            yield 0, entry
        return

    while stack:
        node = stack[-1]
        if node.next_slot >= len(node.slots):
            stack.pop()
            continue

        entry = node.slots[node.next_slot].read_()
        entry_value = entry.value_()

        index = node.index + (node.next_slot << node.shift)
        node.next_slot += 1

        internal = is_internal(node.slots, entry_value)
        if internal > 0:
            stack.append(_XAIteratorNode(to_node(entry_value), index))
        elif internal == 0 and should_yield(entry_value):
            yield index, entry


def xa_is_value(entry: Object) -> bool:
    """
    Return whether an XArray entry is a value.

    See :func:`xa_to_value()`.

    :param entry: ``void *``
    """
    return (entry.value_() & 1) != 0


def xa_to_value(entry: Object) -> Object:
    """
    Return the value in an XArray entry.

    In addition to pointers, XArrays can store integers between 0 and
    ``LONG_MAX``. If :func:`xa_is_value()` returns ``True``, use this to get
    the stored integer.

    >>> entry = xa_load(xa, 9)
    >>> entry
    (void *)0xc9
    >>> xa_is_value(entry)
    True
    >>> xa_to_value(entry)
    (unsigned long)100

    :param entry: ``void *``
    :return: ``unsigned long``
    """
    return cast("unsigned long", entry) >> 1


def xa_is_zero(entry: Object) -> bool:
    """
    Return whether an XArray entry is a "zero" entry.

    A zero entry is an entry that was reserved but is not present. These are
    only visible to the XArray advanced API, so they are only returned by
    :func:`xa_load()` and :func:`xa_for_each()` when ``advanced = True``.

    >>> entry = xa_load(xa, 10, advanced=True)
    >>> entry
    (void *)0x406
    >>> xa_is_zero(entry)
    True
    >>> xa_load(xa, 10)
    (void *)0

    :param entry: ``void *``
    """
    return entry.value_() == _XA_ZERO_ENTRY
